/*
 * CALLBACKS.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module implements functions to collect, display, sort, and deserialize
 * individual data types associated with files that we can enumerate.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

//
//  Sorting support
//

DWORD
SdirCompareAllocationSize (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->AllocationSize, (PULARGE_INTEGER)&Right->AllocationSize);
}

DWORD
SdirCompareCompressedFileSize (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->CompressedFileSize, (PULARGE_INTEGER)&Right->CompressedFileSize);
}

DWORD
SdirCompareCreateDate (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareDate(&Left->CreateTime, &Right->CreateTime);
}

DWORD
SdirCompareCreateTime (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareTime(&Left->CreateTime, &Right->CreateTime);
}

DWORD
SdirCompareFileAttributes (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->FileAttributes < Right->FileAttributes) {
        return SDIR_LESS_THAN;
    } else if (Left->FileAttributes > Right->FileAttributes) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareFileExtension (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareString(Left->Extension, Right->Extension);
}

DWORD
SdirCompareFileId (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FileId, (PULARGE_INTEGER)&Right->FileId);
}

DWORD
SdirCompareFileName (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareString(Left->FileName, Right->FileName);
}

DWORD
SdirCompareFileSize (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FileSize, (PULARGE_INTEGER)&Right->FileSize);
}

DWORD
SdirCompareLinkCount (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->LinkCount < Right->LinkCount) {
        return SDIR_LESS_THAN;
    } else if (Left->LinkCount > Right->LinkCount) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareOsVersion (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->OsVersionHigh < Right->OsVersionHigh) {
        return SDIR_LESS_THAN;
    } else if (Left->OsVersionHigh > Right->OsVersionHigh) {
        return SDIR_GREATER_THAN;
    }

    if (Left->OsVersionLow < Right->OsVersionLow) {
        return SDIR_LESS_THAN;
    } else if (Left->OsVersionLow > Right->OsVersionLow) {
        return SDIR_GREATER_THAN;
    }

    return SDIR_EQUAL;
}

DWORD
SdirCompareReparseTag (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->ReparseTag < Right->ReparseTag) {
        return SDIR_LESS_THAN;
    } else if (Left->ReparseTag > Right->ReparseTag) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareShortName (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareString(Left->ShortFileName, Right->ShortFileName);
}

DWORD
SdirCompareStreamCount (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if (Left->StreamCount < Right->StreamCount) {
        return SDIR_LESS_THAN;
    } else if (Left->StreamCount > Right->StreamCount) {
        return SDIR_GREATER_THAN;
    }
    return SDIR_EQUAL;
}

DWORD
SdirCompareUsn (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->Usn, (PULARGE_INTEGER)&Right->Usn);
}

DWORD
SdirCompareVersion (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareLargeInt((PULARGE_INTEGER)&Left->FileVersion, (PULARGE_INTEGER)&Right->FileVersion);
}

DWORD
SdirCompareWriteDate (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareDate(&Left->WriteTime, &Right->WriteTime);
}

DWORD
SdirCompareWriteTime (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    return SdirCompareTime(&Left->WriteTime, &Right->WriteTime);
}

DWORD
SdirBitwiseFileAttributes (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    if ((Left->FileAttributes & Right->FileAttributes) == Right->FileAttributes) {
        return SDIR_EQUAL;
    }
    return SDIR_NOT_EQUAL;
}

TCHAR
SdirGetUpcasedCharFromString (
    __in LPTSTR Str,
    __in DWORD Index
    )
{
    TCHAR Char = Str[Index];

    if (Char >= 'a' && Char <= 'z') {
        Char = Char - 'a' + 'A';
    }

    return Char;
}

DWORD
SdirBitwiseFileName (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    )
{
    DWORD LeftIndex, RightIndex;

    TCHAR CompareLeft;
    TCHAR CompareRight;

    LeftIndex = 0;
    RightIndex = 0;

    while (Left->FileName[LeftIndex] != '\0' && Right->FileName[RightIndex] != '\0') {

        CompareLeft = SdirGetUpcasedCharFromString(Left->FileName, LeftIndex);
        CompareRight = SdirGetUpcasedCharFromString(Right->FileName, RightIndex);

        LeftIndex++;
        RightIndex++;

        if (CompareRight == '?') {

            //
            //  '?' matches with everything.  We've already advanced to the next
            //  char, so continue.
            //

        } else if (CompareRight == '*') {

            //
            //  Skip one char so Right is the one after *.  Left should compare
            //  the character it's currently on.  Keep going through Left until
            //  we find the char in Right
            //

            LeftIndex--;
            CompareRight = SdirGetUpcasedCharFromString(Right->FileName, RightIndex);
            CompareLeft = SdirGetUpcasedCharFromString(Left->FileName, LeftIndex);

            while (CompareLeft != CompareRight && CompareLeft != '\0') {
                LeftIndex++;
                CompareLeft = SdirGetUpcasedCharFromString(Left->FileName, LeftIndex);
            }

        } else {
            if (CompareLeft != CompareRight) {
                return SDIR_NOT_EQUAL;
            }
        }

    }

    if (Left->FileName[LeftIndex] == '\0' && Right->FileName[RightIndex] == '\0') {
        return SDIR_EQUAL;
    }

    return SDIR_NOT_EQUAL;
}


//
//  File enumeration support
//

BOOL
SdirCollectAllocationSize (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    BOOL RealAllocSize = FALSE;

    if (Opts->GetFileInformationByHandleEx) {

        HANDLE hFile;

        hFile = CreateFile(FullPath,
                           FILE_READ_ATTRIBUTES,
                           FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                           NULL,
                           OPEN_EXISTING,
                           FILE_FLAG_BACKUP_SEMANTICS,
                           NULL);
    
        if (hFile != INVALID_HANDLE_VALUE) {
            FILE_STANDARD_INFO StandardInfo;
    
            if (Opts->GetFileInformationByHandleEx(hFile, FileStandardInfo, &StandardInfo, sizeof(StandardInfo))) {
                Entry->AllocationSize = StandardInfo.AllocationSize;
                RealAllocSize = TRUE;
            }

            CloseHandle(hFile);
        }
    }

    if (!RealAllocSize) {
        ULONG BytesPerSector;
        ULONG SectorsPerCluster;
        ULONG FreeClusters;
        ULONG TotalClusters;
        ULONG ClusterSize;

        GetDiskFreeSpace(Opts->ParentName, &SectorsPerCluster, &BytesPerSector, &FreeClusters, &TotalClusters);

        ClusterSize = SectorsPerCluster * BytesPerSector;

        Entry->AllocationSize.LowPart = FindData->nFileSizeLow;
        Entry->AllocationSize.HighPart = FindData->nFileSizeHigh;

#ifdef SDIR_BIG_SIZE
        Entry->AllocationSize.QuadPart = (Entry->AllocationSize.QuadPart + ClusterSize - 1) & (~(ClusterSize - 1));
#else
        Entry->AllocationSize.LowPart = (Entry->AllocationSize.LowPart + ClusterSize - 1) & (~(ClusterSize - 1));
#endif
    }

    return TRUE;
}

BOOL
SdirCollectCompressedFileSize (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    Entry->CompressedFileSize.LowPart = FindData->nFileSizeLow;
    Entry->CompressedFileSize.HighPart = FindData->nFileSizeHigh;

    if (Opts->GetCompressedFileSize) {
        Entry->CompressedFileSize.LowPart = Opts->GetCompressedFileSize(FullPath, &Entry->CompressedFileSize.HighPart);

        if (Entry->CompressedFileSize.LowPart == INVALID_FILE_SIZE) {
            Entry->CompressedFileSize.LowPart = FindData->nFileSizeLow;
            Entry->CompressedFileSize.HighPart = FindData->nFileSizeHigh;
        }
    }

    return TRUE;
}

BOOL
SdirCollectCreateTime (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    FILETIME tmp;

    FileTimeToLocalFileTime(&FindData->ftCreationTime, &tmp);
    FileTimeToSystemTime(&tmp, &Entry->CreateTime);
    return TRUE;
}

BOOL
SdirCollectFileAttributes (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    DWORD i;

    //
    //  We do this bit by bit to ensure that we don't have file attributes
    //  recorded that we don't understand.  This allows us to perform
    //  equality comparisons where the result is understandable to the user
    //  in that it can be specified and displayed.
    //

    Entry->FileAttributes = 0;
    for (i = 0; i < SdirGetNumAttrPairs(); i++) {
        if (FindData->dwFileAttributes & AttrPairs[i].FileAttribute) {
            Entry->FileAttributes |= AttrPairs[i].FileAttribute;
        }
    }
    return TRUE;
}

BOOL
SdirCollectFileExtension (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    return TRUE;
}

BOOL
SdirCollectFileId (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->FileId.QuadPart = 0;

    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {
        BY_HANDLE_FILE_INFORMATION FileInfo;

        if (GetFileInformationByHandle(hFile, &FileInfo)) {
            Entry->FileId.LowPart = FileInfo.nFileIndexLow;
            Entry->FileId.HighPart = FileInfo.nFileIndexHigh;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectFileName (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    Entry->FileNameLengthInChars = (DWORD)_tcslen(FindData->cFileName);
    if (Entry->FileNameLengthInChars >= MAX_PATH - 1) {
        Entry->FileNameLengthInChars = MAX_PATH - 2;
    }
    if (Entry->FileNameLengthInChars > SdirDirCollectionLongest) {
        SdirDirCollectionLongest = Entry->FileNameLengthInChars;
    }

    SdirDirCollectionTotalNameLength += Entry->FileNameLengthInChars;

    memcpy(Entry->FileName, FindData->cFileName, Entry->FileNameLengthInChars * sizeof(TCHAR));
    Entry->FileName[Entry->FileNameLengthInChars] = '\0';

    Entry->Extension = _tcsrchr(Entry->FileName, '.');

    //
    //  For simplicity's sake, if we have no extension set the field
    //  to the end of string, so we'll see a valid pointer of nothing.
    //

    if (Entry->Extension == NULL) {
        Entry->Extension = Entry->FileName + Entry->FileNameLengthInChars;
    } else {
        Entry->Extension++;
    }

    return TRUE;
}

BOOL
SdirCollectFileSize (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    Entry->FileSize.LowPart = FindData->nFileSizeLow;
    Entry->FileSize.HighPart = FindData->nFileSizeHigh;
    return TRUE;
}

BOOL
SdirCollectLinkCount (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->LinkCount = 0;

    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {
        BY_HANDLE_FILE_INFORMATION FileInfo;

        if (GetFileInformationByHandle(hFile, &FileInfo)) {
            Entry->LinkCount = FileInfo.nNumberOfLinks;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectOsVersion (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFileRead;
    union {
        IMAGE_DOS_HEADER DosHeader;
        struct {
            DWORD Signature;
            IMAGE_FILE_HEADER ImageHeader;
            IMAGE_OPTIONAL_HEADER OptionalHeader;
        } s2;
    } u1;
    DWORD BytesReturned;

    Entry->OsVersionHigh = 0;
    Entry->OsVersionLow = 0;

    //
    //  We want the earlier handle to be attribute only so we can
    //  operate on directories, but we need data for this, so we
    //  end up with two handles.
    //

    hFileRead = CreateFile(FullPath,
                           FILE_READ_ATTRIBUTES|FILE_READ_DATA,
                           FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                           NULL,
                           OPEN_EXISTING,
                           FILE_FLAG_BACKUP_SEMANTICS,
                           NULL);

    if (hFileRead != INVALID_HANDLE_VALUE) {

        if (ReadFile(hFileRead, &u1.DosHeader, sizeof(u1.DosHeader), &BytesReturned, NULL) &&
            BytesReturned == sizeof(u1.DosHeader) &&
            u1.DosHeader.e_magic == IMAGE_DOS_SIGNATURE &&
            u1.DosHeader.e_lfanew != 0) {

            SetFilePointer(hFileRead, u1.DosHeader.e_lfanew, NULL, FILE_BEGIN);

            if (ReadFile(hFileRead, &u1.s2, sizeof(u1.s2), &BytesReturned, NULL) &&
                BytesReturned == sizeof(u1.s2) &&
                u1.s2.Signature == IMAGE_NT_SIGNATURE &&
                u1.s2.ImageHeader.SizeOfOptionalHeader >= FIELD_OFFSET(IMAGE_OPTIONAL_HEADER, Subsystem)) {

                Entry->OsVersionHigh = u1.s2.OptionalHeader.MajorSubsystemVersion;
                Entry->OsVersionLow = u1.s2.OptionalHeader.MinorSubsystemVersion;
            }
        }
        CloseHandle(hFileRead);
    }
    return TRUE;
}

BOOL
SdirCollectReparseTag (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    if (Entry->FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) {
        Entry->ReparseTag = FindData->dwReserved0;
    } else {
        Entry->ReparseTag = 0;
    }
    return TRUE;
}

BOOL
SdirCollectShortName (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    if (FindData->cAlternateFileName[0] == '\0' && _tcslen(FindData->cFileName) <= 12) {
        CopyMemory(Entry->ShortFileName, FindData->cFileName, sizeof(FindData->cAlternateFileName));
    } else {
        CopyMemory(Entry->ShortFileName, FindData->cAlternateFileName, sizeof(FindData->cAlternateFileName));
    }
    return TRUE;
}

#ifdef UNICODE
BOOL
SdirCollectStreamCount (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFind;
    SDIR_WIN32_FIND_STREAM_DATA FindStreamData;

    Entry->StreamCount = 0;

    //
    //  These APIs are Unicode only.  We could do an ANSI to Unicode thunk here, but
    //  since Unicode is the default build and ANSI is only useful for older systems
    //  (where this API won't exist) there doesn't seem much point.
    //

    if (Opts->FindFirstStreamW && Opts->FindNextStreamW) {
        hFind = Opts->FindFirstStreamW(FullPath, 0, &FindStreamData, 0);
        if (hFind != INVALID_HANDLE_VALUE) {

            do {
                Entry->StreamCount++;
            } while (Opts->FindNextStreamW(hFind, &FindStreamData));
        }

        FindClose(hFind);
    }

    return TRUE;
}
#endif

BOOL
SdirCollectUsn (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    HANDLE hFile;

    Entry->Usn.QuadPart = 0;
    hFile = CreateFile(FullPath,
                       FILE_READ_ATTRIBUTES,
                       FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_BACKUP_SEMANTICS,
                       NULL);

    if (hFile != INVALID_HANDLE_VALUE) {

        struct {
            SDIR_USN_RECORD UsnRecord;
            WCHAR FileName[MAX_PATH];
        } s1;
        DWORD BytesReturned;

        if (DeviceIoControl(hFile, FSCTL_READ_FILE_USN_DATA, NULL, 0, &s1, sizeof(s1), &BytesReturned, NULL)) {
            Entry->Usn.QuadPart = s1.UsnRecord.Usn.QuadPart;
        }

        CloseHandle(hFile);
    }
    return TRUE;
}

BOOL
SdirCollectVersion (
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    DWORD Junk;
    PVOID Buffer;
    DWORD VerSize = GetFileVersionInfoSize(FullPath, &Junk);
    VS_FIXEDFILEINFO * RootBlock;

    Entry->FileVersion.QuadPart = 0;
    Entry->FileVersionFlags = 0;

    Buffer = HeapAlloc(GetProcessHeap(), 0, VerSize);
    if (Buffer != NULL) {
        if (GetFileVersionInfo(FullPath, 0, VerSize, Buffer)) {
            if (VerQueryValue(Buffer, _T("\\"), &RootBlock, &Junk)) {
                Entry->FileVersion.HighPart = RootBlock->dwFileVersionMS;
                Entry->FileVersion.LowPart = RootBlock->dwFileVersionLS;
                Entry->FileVersionFlags = RootBlock->dwFileFlags & RootBlock->dwFileFlagsMask;
            }
        }
        HeapFree(GetProcessHeap(), 0, Buffer);
    }
    return TRUE;
}

BOOL
SdirCollectWriteTime(
    __inout PSDIR_DIRENT Entry,
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    )
{
    FILETIME tmp;

    FileTimeToLocalFileTime(&FindData->ftLastWriteTime, &tmp);
    FileTimeToSystemTime(&tmp, &Entry->WriteTime);
    return TRUE;
}

BOOL
SdirCollectSummary(
    __in PSDIR_DIRENT Entry
    )
{
    if (Entry->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
        Summary->NumDirs++;
    } else {
        Summary->NumFiles++;
    }

#if SDIR_BIG_SIZE
    Summary->TotalSize += Entry->FileSize.QuadPart;
#else
    Summary->TotalSize += Entry->FileSize.LowPart;
#endif

    if (Opts->EnableCompressedFileSize) {
#if SDIR_BIG_SIZE
        Summary->CompressedSize += Entry->CompressedFileSize.QuadPart;
#else
        Summary->CompressedSize += Entry->CompressedFileSize.LowPart;
#endif
    }

    return TRUE;
}

//
//  When criteria are specified to apply attributes, we need to load the specification
//  into a dummy dirent to perform comparisons against.  The below functions implement
//  these.
//

BOOL
SdirGenerateAllocationSize(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->AllocationSize = SdirStringToFileSize(String);
    return TRUE;
}

BOOL
SdirGenerateCompressedFileSize(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->CompressedFileSize = SdirStringToFileSize(String);
    return TRUE;
}

BOOL
SdirGenerateCreateDate(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    return SdirStringToDate(String, &Entry->CreateTime);
}

BOOL
SdirGenerateCreateTime(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    return SdirStringToTime(String, &Entry->CreateTime);
}

BOOL
SdirGenerateFileAttributes(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    DWORD i;

    Entry->FileAttributes = 0;

    while (*String) {

        for (i = 0; i < SdirGetNumAttrPairs(); i++) {
            if (*String == AttrPairs[i].DisplayLetter) {
                Entry->FileAttributes |= AttrPairs[i].FileAttribute;
            }
        }

        String++;
    }
    return TRUE;
}

BOOL
SdirGenerateFileExtension(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    //
    //  Since we have one dirent per comparison, just shove the extension in
    //  the file name buffer and point the extension to it.  This buffer can't
    //  be used for anything else anyway.
    //

    _stprintf_s(Entry->FileName, sizeof(Entry->FileName)/sizeof(Entry->FileName[0]), _T("%s"), String);
    Entry->FileName[sizeof(Entry->FileName)/sizeof(Entry->FileName[0]) - 1] = '\0';
    Entry->Extension = Entry->FileName;

    return TRUE;
}

BOOL
SdirGenerateFileName(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->FileNameLengthInChars = _stprintf_s(Entry->FileName, sizeof(Entry->FileName)/sizeof(Entry->FileName[0]), _T("%s"), String);
    Entry->FileName[sizeof(Entry->FileName)/sizeof(Entry->FileName[0]) - 1] = '\0';

    return TRUE;
}

BOOL
SdirGenerateFileSize(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->FileSize = SdirStringToFileSize(String);
    return TRUE;
}

BOOL
SdirGenerateLinkCount(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->LinkCount = SdirStringToNum32(String, NULL);
    return TRUE;
}

BOOL
SdirGenerateOsVersion(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    TCHAR * endptr = NULL;

    Entry->OsVersionHigh = (WORD)SdirStringToNum32(String, &endptr);

    if (*endptr == '.') {
        Entry->OsVersionLow = (WORD)SdirStringToNum32(endptr+1, NULL);
    }

    return TRUE;
}

BOOL
SdirGenerateReparseTag(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->ReparseTag = SdirStringToNum32(String, NULL);
    return TRUE;
}

BOOL
SdirGenerateShortName(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    _stprintf_s(Entry->ShortFileName, sizeof(Entry->ShortFileName)/sizeof(Entry->ShortFileName[0]), _T("%s"), String);
    Entry->ShortFileName[sizeof(Entry->ShortFileName)/sizeof(Entry->ShortFileName[0]) - 1] = '\0';

    return TRUE;
}

BOOL
SdirGenerateStreamCount(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->StreamCount = SdirStringToNum32(String, NULL);
    return TRUE;
}

BOOL
SdirGenerateUsn(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    Entry->Usn = SdirStringToNum64(String, NULL);
    return TRUE;
}

BOOL
SdirGenerateVersion(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    LARGE_INTEGER Version = {0};
    TCHAR * endptr = NULL;

    Version.HighPart = SdirStringToNum32(String, &endptr)<<16;

    if (*endptr == '.') {
        Version.HighPart = Version.HighPart + (WORD)SdirStringToNum32(endptr+1, &endptr);

        if (*endptr == '.') {
            Version.LowPart = SdirStringToNum32(endptr+1, &endptr)<<16;
            if (*endptr == '.') {
                Version.LowPart = Version.LowPart + (WORD)SdirStringToNum32(endptr+1, NULL);
            }
                
        }
    }

    Entry->FileVersion = Version;

    return TRUE;
}

BOOL
SdirGenerateWriteDate(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    return SdirStringToDate(String, &Entry->WriteTime);
}

BOOL
SdirGenerateWriteTime(
    __inout PSDIR_DIRENT Entry,
    __in LPTSTR String
    )
{
    return SdirStringToTime(String, &Entry->WriteTime);
}


//
//  Specific formatting and sizing callbacks for each supported piece of file metadata.
//

ULONG
SdirDisplayAllocationSize (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayGenericSize(Buffer, SDIR_FILE_SIZE_COLOR, &Entry->AllocationSize);
}

ULONG
SdirDisplayCompressedFileSize (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayGenericSize(Buffer, SDIR_FILE_SIZE_COLOR, &Entry->CompressedFileSize);
}

ULONG
SdirDisplayFileCreateDate (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileDate(Buffer, DefaultAttributes, &Entry->CreateTime);
}

ULONG
SdirDisplayFileCreateTime (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileTime(Buffer, DefaultAttributes, &Entry->CreateTime);
}

ULONG
SdirDisplayFileAttributes (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR StrAtts[32];
    DWORD i;

    if (Buffer) {
        StrAtts[0] = ' ';
        for (i = 0; i < SdirGetNumAttrPairs(); i++) {
            if (Entry->FileAttributes & AttrPairs[i].FileAttribute) {
                StrAtts[i + 1] = AttrPairs[i].DisplayLetter;
            } else {
                StrAtts[i + 1] = '-';
            }
        }

        StrAtts[i + 1] = '\0';
    
        SdirPasteStr(Buffer, StrAtts, SDIR_FILE_SIZE_COLOR, SdirGetNumAttrPairs() + 1);
    }
    return SdirGetNumAttrPairs() + 1;
}

ULONG
SdirDisplayFileId (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayHex64(Buffer, DefaultAttributes, &Entry->FileId);
}

ULONG
SdirDisplayFileSize (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    if (Buffer) {
        if (Entry->FileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) {
            SdirPasteStr(Buffer, _T(" <LNK>"), DefaultAttributes, 6);
            return 6;
        } else if (Entry->FileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
            SdirPasteStr(Buffer, _T(" <DIR>"), DefaultAttributes, 6);
            return 6;
        } else {
            return SdirDisplayGenericSize(Buffer, SDIR_FILE_SIZE_COLOR, &Entry->FileSize);
        }
    }
    return 6;
}

ULONG
SdirDisplayLinkCount (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[5];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %03i"), Entry->LinkCount);
        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 4);
    }
    return 4;
}

ULONG
SdirDisplayOsVersion (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[30];
    BYTE  ThisOsMajor = LOBYTE(LOWORD(Opts->OsVersion));
    BYTE  ThisOsMinor = HIBYTE(LOWORD(Opts->OsVersion));
    WORD  Attributes = SDIR_FILE_TIME_COLOR;

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %02i.%02i"), Entry->OsVersionHigh, Entry->OsVersionLow);
    
        if (Entry->OsVersionHigh > ThisOsMajor ||
            Entry->OsVersionHigh == ThisOsMajor && Entry->OsVersionLow > ThisOsMinor) {
    
            Attributes = SDIR_FUTURE_VERSION_COLOR;
        }
    
        SdirPasteStr(Buffer, Str, Attributes, 6);
    }
    return 6;
}

ULONG
SdirDisplayReparseTag (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    ULONG Tag = 0;
    if (Entry) {
        Tag = Entry->ReparseTag;
    }
    return SdirDisplayHex32(Buffer, DefaultAttributes, Tag);
}

ULONG
SdirDisplayShortName (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    ULONG CurrentChar = 0;
    DWORD ShortLength;

    //
    //  Just because we like special cases, add a space only if both
    //  names are being displayed.
    //

    if (Opts->EnableFileName & SDIR_OPT_DISPLAY) {
        if (Buffer) {
            SdirPasteStrAndPad(&Buffer[CurrentChar], NULL, 0, 0, 1);
        }
        CurrentChar++;
    }

    if (Buffer) {
        ShortLength = (DWORD)_tcslen(Entry->ShortFileName);

        SdirPasteStrAndPad(&Buffer[CurrentChar], Entry->ShortFileName, DefaultAttributes, ShortLength, 12);
    }
    CurrentChar += 12;
    return CurrentChar;
}

ULONG
SdirDisplayStreamCount (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[5];

    if (Buffer) {
        _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %03i"), Entry->StreamCount);
        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 4);
    }
    return 4;
}

ULONG
SdirDisplayUsn(
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayHex64(Buffer, DefaultAttributes, &Entry->Usn);
}

ULONG
SdirDisplayVersion(
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    TCHAR Str[40];

    if (Buffer) {
        _stprintf_s(Str,
                    sizeof(Str)/sizeof(Str[0]),
                    _T(" %02i.%02i.%05i.%05i %c%c"),
                    Entry->FileVersion.HighPart >> 16,
                    Entry->FileVersion.HighPart & 0xffff,
                    Entry->FileVersion.LowPart >> 16,
                    Entry->FileVersion.LowPart & 0xffff,
                    Entry->FileVersionFlags & VS_FF_DEBUG ? 'D' : '-',
                    Entry->FileVersionFlags & VS_FF_PRERELEASE ? 'P' : '-');

        SdirPasteStr(Buffer, Str, SDIR_FILE_TIME_COLOR, 21);
    }
    return 21;
}

ULONG
SdirDisplaySummary(
    __in WORD DefaultAttributes
    )
{
    SDIR_FMTCHAR Buffer[200];
    TCHAR Str[100];
    DWORD Len;
    LARGE_INTEGER Size;
    ULONG CurrentChar = 0;

    _stprintf_s(Str, sizeof(Str)/sizeof(Str[0]), _T(" %i files, %i dirs,"), Summary->NumFiles, Summary->NumDirs);
    Len = (DWORD)_tcslen(Str);
    SdirPasteStr(&Buffer[CurrentChar], Str, DefaultAttributes, Len);
    CurrentChar += Len;

#if SDIR_BIG_SIZE
    Size.QuadPart = Summary->TotalSize;
#else
    Size.HighPart = 0;
    Size.LowPart = Summary->TotalSize;
#endif

    CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], DefaultAttributes, &Size);
    SdirPasteStr(&Buffer[CurrentChar], _T(" used,"), DefaultAttributes, 6);
    CurrentChar += 6;

    if (Opts->EnableCompressedFileSize) {
#if SDIR_BIG_SIZE
        Size.QuadPart = Summary->CompressedSize;
#else
        Size.HighPart = 0;
        Size.LowPart = Summary->CompressedSize;
#endif

        CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], DefaultAttributes, &Size);
        SdirPasteStr(&Buffer[CurrentChar], _T(" compressed,"), DefaultAttributes, 12);
        CurrentChar += 12;
    }

    CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], DefaultAttributes, &Summary->VolumeSize);
    SdirPasteStr(&Buffer[CurrentChar], _T(" vol size,"), DefaultAttributes, 10);
    CurrentChar += 10;

    CurrentChar += SdirDisplayGenericSize(&Buffer[CurrentChar], DefaultAttributes, &Summary->FreeSize);
    SdirPasteStr(&Buffer[CurrentChar], _T(" vol free"), DefaultAttributes, 9);
    CurrentChar += 9;

    SdirWrite(Buffer, CurrentChar);

    return CurrentChar;
}
        

ULONG
SdirDisplayFileWriteDate (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileDate(Buffer, DefaultAttributes, &Entry->WriteTime);
}

ULONG
SdirDisplayFileWriteTime (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    )
{
    return SdirDisplayFileTime(Buffer, DefaultAttributes, &Entry->WriteTime);
}

//
//  The below table corresponds to the supported options in the program.
//  Each option has a flags value describing whether to collect or display
//  the piece of metadata, a default, a switch to turn display on or off
//  or sort by it, a callback function to tell how much space it will need,
//  an optional sort compare function, and some help text.
//
//  File names, extensions and attributes are always collected, because
//  that's how we determine colors.
//

#define OPT_OS(x) FIELD_OFFSET(SDIR_OPTS, x)

const SDIR_OPT SdirOptions[] = {
    {OPT_OS(EnableAllocationSize),     _T("as"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayAllocationSize,     SdirCollectAllocationSize,
        SdirCompareAllocationSize,     NULL,                          SdirGenerateAllocationSize,
        "allocation size"},
    
    {OPT_OS(EnableCreateDate),         _T("cd"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayFileCreateDate,     SdirCollectCreateTime,
        SdirCompareCreateDate,         NULL,                          SdirGenerateCreateDate,
        "create date"},

    {OPT_OS(EnableCompressedFileSize), _T("cs"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayCompressedFileSize, SdirCollectCompressedFileSize,
        SdirCompareCompressedFileSize, NULL,                          SdirGenerateCompressedFileSize,
        "compressed size"},

    {OPT_OS(EnableCreateTime),         _T("ct"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayFileCreateTime,     SdirCollectCreateTime,
        SdirCompareCreateTime,         NULL,                          SdirGenerateCreateTime,
        "create time"},

    {OPT_OS(EnableFileAttributes),     _T("fa"), SDIR_OPT_COLLECT|SDIR_OPT_ALLOW_DISPLAY,
        SdirDisplayFileAttributes,     SdirCollectFileAttributes,
        SdirCompareFileAttributes,     SdirBitwiseFileAttributes,     SdirGenerateFileAttributes,
        "file attributes"},

    {OPT_OS(EnableFileExtension),      _T("fe"), SDIR_OPT_DISPLAY|SDIR_OPT_COLLECT|SDIR_OPT_ALLOW_SORT,
        NULL,                          SdirCollectFileExtension,
        SdirCompareFileExtension,      NULL,                          SdirGenerateFileExtension,
        "file extension"},

    {OPT_OS(EnableFileId),             _T("fi"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayFileId,             SdirCollectFileId,
        SdirCompareFileId,             NULL,                          NULL,
        "file id"},

    {OPT_OS(EnableFileName),           _T("fn"), SDIR_OPT_DISPLAY|SDIR_OPT_COLLECT|SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        NULL,                          SdirCollectFileName,
        SdirCompareFileName,           SdirBitwiseFileName,           SdirGenerateFileName,
        "file name"},

    {OPT_OS(EnableFileSize),           _T("fs"), SDIR_OPT_DISPLAY|SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayFileSize,           SdirCollectFileSize,
        SdirCompareFileSize,           NULL,                          SdirGenerateFileSize,
        "file size"},

    {OPT_OS(EnableLinkCount),          _T("lc"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayLinkCount,          SdirCollectLinkCount,
        SdirCompareLinkCount,          NULL,                          SdirGenerateLinkCount,
        "link count"},

#ifdef UNICODE
    {OPT_OS(EnableNamedStreams),       _T("ns"), SDIR_OPT_ALLOW_DISPLAY,
        NULL,                          NULL,                   
        NULL,                          NULL,                          NULL,
        "named streams"},
#endif

    {OPT_OS(EnableOsVersion),          _T("os"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayOsVersion,          SdirCollectOsVersion,
        SdirCompareOsVersion,          NULL,                          SdirGenerateOsVersion,
        "minimum OS version"},

    {OPT_OS(EnableReparseTag),         _T("rt"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayReparseTag,         SdirCollectReparseTag,
        SdirCompareReparseTag,         NULL,                          SdirGenerateReparseTag,
        "reparse tag"},

#ifdef UNICODE
    {OPT_OS(EnableStreamCount),        _T("sc"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayStreamCount,        SdirCollectStreamCount,
        SdirCompareStreamCount,        NULL,                          SdirGenerateStreamCount,
        "stream count"},
#endif

    {OPT_OS(EnableSummary),            _T("sm"), SDIR_OPT_DISPLAY|SDIR_OPT_COLLECT|SDIR_OPT_ALLOW_DISPLAY,
        NULL,                          NULL,
        NULL,                          NULL,                          NULL,
        "summary"},

    {OPT_OS(EnableShortName),          _T("sn"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayShortName,          SdirCollectShortName,
        SdirCompareShortName,          NULL,                          SdirGenerateShortName,
        "short name"},

    {OPT_OS(EnableUsn),                _T("us"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayUsn,                SdirCollectUsn,
        SdirCompareUsn,                NULL,                          SdirGenerateUsn,
        "USN"},

    {OPT_OS(EnableVersion),            _T("vr"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayVersion,            SdirCollectVersion,
        SdirCompareVersion,            NULL,                          SdirGenerateVersion,
        "version"},

    {OPT_OS(EnableWriteDate),          _T("wd"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayFileWriteDate,      SdirCollectWriteTime,
        SdirCompareWriteDate,          NULL,                          SdirGenerateWriteDate,
        "write date"},

    {OPT_OS(EnableWriteTime),          _T("wt"), SDIR_OPT_ALLOW_DISPLAY|SDIR_OPT_ALLOW_SORT,
        SdirDisplayFileWriteTime,      SdirCollectWriteTime,
        SdirCompareWriteTime,          NULL,                          SdirGenerateWriteTime,
        "write time"},
};

DWORD
SdirGetNumSdirOptions()
{
    return sizeof(SdirOptions)/sizeof(SdirOptions[0]);
}

//
//  This table is really quite redundant, but it corresponds to the order
//  in which we will render each piece of metadata.  The table above is
//  the order we display help.
//


const SDIR_EXEC SdirExec[] = {
    {OPT_OS(EnableFileId),             SdirDisplayFileId},
    {OPT_OS(EnableUsn),                SdirDisplayUsn},
    {OPT_OS(EnableFileSize),           SdirDisplayFileSize},
    {OPT_OS(EnableCompressedFileSize), SdirDisplayCompressedFileSize},
    {OPT_OS(EnableAllocationSize),     SdirDisplayAllocationSize},
    {OPT_OS(EnableFileAttributes),     SdirDisplayFileAttributes},
    {OPT_OS(EnableReparseTag),         SdirDisplayReparseTag},
    {OPT_OS(EnableLinkCount),          SdirDisplayLinkCount},
    {OPT_OS(EnableStreamCount),        SdirDisplayStreamCount},
    {OPT_OS(EnableCreateDate),         SdirDisplayFileCreateDate},
    {OPT_OS(EnableCreateTime),         SdirDisplayFileCreateTime},
    {OPT_OS(EnableWriteDate),          SdirDisplayFileWriteDate},
    {OPT_OS(EnableWriteTime),          SdirDisplayFileWriteTime},
    {OPT_OS(EnableVersion),            SdirDisplayVersion},
    {OPT_OS(EnableOsVersion),          SdirDisplayOsVersion},
};

DWORD
SdirGetNumSdirExec()
{
    return sizeof(SdirExec)/sizeof(SdirExec[0]);
}

// vim:sw=4:ts=4:et:
