/*
 * COLOR.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module implements string parsing and rule application to select a
 * given set of color attributes to render any particular file with.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

SDIR_ATTRIBUTE_COLOR_STRING ColorString[] = {

    //
    //  Some CGA color constants
    //

    {_T("black"),        0x00},
    {_T("blue"),         0x01},
    {_T("green"),        0x02},
    {_T("cyan"),         0x03},
    {_T("red"),          0x04},
    {_T("magenta"),      0x05},
    {_T("brown"),        0x06},
    {_T("gray"),         0x07},

    {_T("darkgray"),     0x08},
    {_T("lightblue"),    0x09},
    {_T("lightgreen"),   0x0A},
    {_T("lightcyan"),    0x0B},
    {_T("lightred"),     0x0C},
    {_T("lightmagenta"), 0x0D},
    {_T("yellow"),       0x0E},
    {_T("white"),        0x0F},

    //
    //  Helper modifier
    //

    {_T("bright"),       0x08},

    //
    //  Some non-color things
    //

    {_T("invert"),       0x100},
    {_T("hide"),         0x200},
    {_T("continue"),     0x400}

};


WORD
SdirAttributeFromString(
    __in LPTSTR String
    )
{
    WORD Attribute = 0;
    DWORD Element;
    BOOL Background = FALSE;
    LPTSTR Next;

    //
    //  If it's a number, parse it like we used to.
    //

    if (String[0] >= '0' && String[0] <= '9') {
        return (WORD)SdirStringToNum32(String, NULL);
    }

    //
    //  Loop through the list of color keywords
    //

    while (String && String[0] != '\0') {

        //
        //  Check if we're setting a background.
        //
    
        Background = FALSE;
        if ((String[0] == 'b' || String[0] == 'B') &&
            (String[1] == 'g' || String[1] == 'G') &&
            (String[2] == '_' || String[2] == '_')) {
    
            String += 3;
            Background = TRUE;
        }

        //
        //  Look for the next element
        //

        Next = _tcschr(String, '+');
        if (Next) {
            *Next = '\0';
        }

        //
        //  Walk through the string table for a match.
        //
    
        for (Element = 0; Element < sizeof(ColorString)/sizeof(ColorString[0]); Element++) {
            if (_tcsicmp(String, ColorString[Element].String) == 0) {
    
                if (Background) {
                    Attribute |= (ColorString[Element].Attribute & SDIR_ATTRIBUTE_COLOR_MASK) << 4;
                } else {
                    Attribute |= ColorString[Element].Attribute;
                }
    
                break;
            }
        }
    
        //
        //  Advance to the next element
        //

        if (Next) {
            String = Next + 1;
        } else {
            String = NULL;
        }
    }


    return Attribute;
}

typedef struct _SDIR_ATTRIBUTE_APPLY {
    SDIR_COMPARE_FN CompareFn;
    BOOL TruthStates[3];
    WORD Attribute;
    SDIR_DIRENT CompareEntry;
} SDIR_ATTRIBUTE_APPLY, *PSDIR_ATTRIBUTE_APPLY;


PSDIR_ATTRIBUTE_APPLY SdirAttributeApply = NULL; 
const
CHAR SdirAttributeDefaultApplyString[] = 
    "fa&r,magenta;"
    "fa&D,lightmagenta;"
    "fa&R,green;"
    "fa&H,green;"
    "fa&S,green;"
    "fe=bat,lightred;"
    "fe=cmd,lightred;"
    "fe=com,lightcyan;"
    "fe=dll,cyan;"
    "fe=doc,white;"
    "fe=docx,white;"
    "fe=exe,lightcyan;"
    "fe=htm,white;"
    "fe=html,white;"
    "fe=pdf,white;"
    "fe=pl,red;"
    "fe=ppt,white;"
    "fe=pptx,white;"
    "fe=sys,cyan;"
    "fe=xls,white;"
    "fe=xlsx,white";

BOOL
SdirParseAttributeApplyString()
{
    TCHAR SingleElement[256];
    TCHAR SingleSwitch[20];
    LPTSTR Next;
    LPTSTR This;
    LPTSTR Operator;
    LPTSTR Criteria;
    LPTSTR Attribute;
    ULONG i;
    PSDIR_ATTRIBUTE_APPLY ThisApply;
    LPTSTR SdirAttributeCombinedApplyString = NULL;
    DWORD  SdirColorPrependLength = 0;
    DWORD  SdirColorAppendLength  = 0;
    DWORD  SdirColorReplaceLength = 0;
    PUCHAR OptEnabled;

    //
    //  Load any user specified colors from the environment.  Prepend values go before
    //  any default; replace values supersede any default; and append values go last.
    //  We need to insert semicolons between these values if they're specified.
    //
    //  First, count how big the allocation needs to be and allocate it.
    //
    //  TODO - Should we support these coming from the command line (really SDIR_OPTS?)
    //

    SdirColorPrependLength = GetEnvironmentVariable(_T("SDIR_COLOR_PREPEND"), NULL, 0);
    SdirColorReplaceLength = GetEnvironmentVariable(_T("SDIR_COLOR_REPLACE"), NULL, 0);
    SdirColorAppendLength  = GetEnvironmentVariable(_T("SDIR_COLOR_APPEND"), NULL, 0);

    SdirAttributeCombinedApplyString = HeapAlloc(GetProcessHeap(), 0, (SdirColorPrependLength + 1 + SdirColorAppendLength + 1 + SdirColorReplaceLength + sizeof(SdirAttributeDefaultApplyString)) * sizeof(TCHAR));

    if (SdirAttributeCombinedApplyString == NULL) {
        return FALSE;
    }

    //
    //  Now, load any environment variables into the buffer.  If replace isn't specified,
    //  we use the hardcoded default.
    //

    if (SdirColorPrependLength) {
        GetEnvironmentVariable(_T("SDIR_COLOR_PREPEND"), SdirAttributeCombinedApplyString, SdirColorPrependLength);
        SdirAttributeCombinedApplyString[SdirColorPrependLength - 1] = ';';
    }

    if (SdirColorReplaceLength) {
        GetEnvironmentVariable(_T("SDIR_COLOR_REPLACE"), SdirAttributeCombinedApplyString + SdirColorPrependLength, SdirColorReplaceLength);
        SdirColorReplaceLength--;
    } else {
        _stprintf_s(SdirAttributeCombinedApplyString + SdirColorPrependLength, 
                    sizeof(SdirAttributeDefaultApplyString) / sizeof(SdirAttributeDefaultApplyString[0]),
                    _T("%hs"),
                    SdirAttributeDefaultApplyString);
        SdirColorReplaceLength = sizeof(SdirAttributeDefaultApplyString) / sizeof(SdirAttributeDefaultApplyString[0]) - 1;
    }

    if (SdirColorAppendLength) {
        SdirAttributeCombinedApplyString[SdirColorPrependLength + SdirColorReplaceLength] = ';';
        SdirColorReplaceLength++;
        GetEnvironmentVariable(_T("SDIR_COLOR_APPEND"), SdirAttributeCombinedApplyString + SdirColorPrependLength + SdirColorReplaceLength, SdirColorAppendLength);
    }

    //
    //  First count the number of rules so we can allocate the correct amount
    //  of memory for the set.  For ease of use, we allocate an additional one
    //  as a "null terminator."
    //

    This = SdirAttributeCombinedApplyString;
    i = 1;

    while (This != NULL) {

        i++;
        Next = _tcschr(This, ';');

        if (Next) {
            This = Next + 1;
        } else {
            This = NULL;
        }
    }

    SdirAttributeApply = HeapAlloc(GetProcessHeap(), 0, i * sizeof(SDIR_ATTRIBUTE_APPLY));
    if (SdirAttributeApply == NULL) {
        HeapFree(GetProcessHeap(), 0, SdirAttributeCombinedApplyString);
        return FALSE;
    }

    ZeroMemory(SdirAttributeApply, i * sizeof(SDIR_ATTRIBUTE_APPLY));
    ThisApply = SdirAttributeApply;
    This = SdirAttributeCombinedApplyString;

    //
    //  Now go through the rules again and populate the memory we just
    //  allocated with the comparison functions, criteria and attributes.
    //

    while (This != NULL) {

        //
        //  We have a mutable buffer since we just allocated it.
        //  NULL-ing as we go is really a hint to new 'secure'
        //  versions of the CRT that we don't intend to copy more
        //  TCHARs than this, so it doesn't need to gracelessly
        //  puke.
        //

        Next = _tcschr(This, ';');
        if (Next) {
            *Next = '\0';
        }

        //
        //  Ignore any leading spaces
        //

        while(*This == ' ') {
            This++;
        }

        //
        //  We want to copy as many tchars as there are between two
        //  elements, but not more than the buffer size.  Note that
        //  the printf routine below will do the right thing at end
        //  of string so it's okay to pass a larger buffer size
        //  there.
        //

        i = sizeof(SingleElement)/sizeof(SingleElement[0]);

        if (Next && ((ULONG)(Next - This) < i)) {
            i = Next - This + 1;
        }

        //
        //  If this element is empty, ignore it
        //

        if (i == 0 || *This == '\0') {
            goto next_element;
        }

        _stprintf_s(SingleElement, i, _T("%s"), This);
        SingleElement[i - 1] = '\0';

        //
        //  Count the number of chars to the operator.  Copy that
        //  many into the switch buffer, but not more than the
        //  size of the buffer.  We increment i to ensure we have
        //  space for the null terminator as well as the
        //  substring we want to process.
        //

        i = _tcscspn(SingleElement, _T("&<>=!"));
        if (i >= sizeof(SingleSwitch)/sizeof(SingleSwitch[0])) {
            i = sizeof(SingleSwitch)/sizeof(SingleSwitch[0]);
        } else {
            i++;
        }

        //
        //  Here we just give up on the 'secure' CRT things.  We
        //  want to copy the switch, up to the size of the buffer,
        //  and we're doing it manually to avoid puking because
        //  the source string may still (will still) have characters
        //  beyond this number.
        //

        memcpy(SingleSwitch, SingleElement, i * sizeof(TCHAR));
        SingleSwitch[i - 1] = '\0';

        Operator = SingleElement + i - 1;

        //
        //  Count the length of the operator and verify it's
        //  sane.
        //

        i = _tcsspn(Operator, _T("&<>=!"));

        if (i != 1 && i != 2) {
            SdirWriteString(_T("Operator error in: "));
            goto error_return;
        }

        //
        //  Count the length of the criteria and null terminate
        //  it for convenience.
        //

        Criteria = Operator + i;

        i = _tcscspn(Criteria, _T(","));
        if (Criteria[i] == '\0') {
            SdirWriteString(_T("Attribute error in: "));
            goto error_return;
        }

        Criteria[i] = '\0';
        Attribute = Criteria + i + 1;

        for (i = 0; i < SdirGetNumSdirOptions(); i++) {
            if (_tcsicmp(SingleSwitch, SdirOptions[i].Switch) == 0 &&
                SdirOptions[i].GenerateFromStringFn != NULL) {

                //
                //  Based on the operator, fill in the truth table.  We'll
                //  use the generic compare function and based on this
                //  truth table we'll decide whether the rule is satisfied
                //  or not.
                //

                if (Operator[0] == '>') {
                    ThisApply->CompareFn = SdirOptions[i].CompareFn;

                    ThisApply->TruthStates[SDIR_LESS_THAN] = FALSE;
                    ThisApply->TruthStates[SDIR_GREATER_THAN] = TRUE;
                    if (Operator[1] == '=') {
                        ThisApply->TruthStates[SDIR_EQUAL] = TRUE;
                    } else {
                        ThisApply->TruthStates[SDIR_EQUAL] = FALSE;
                    }
                } else if (Operator[0] == '<') {
                    ThisApply->CompareFn = SdirOptions[i].CompareFn;

                    ThisApply->TruthStates[SDIR_LESS_THAN] = TRUE;
                    ThisApply->TruthStates[SDIR_GREATER_THAN] = FALSE;
                    if (Operator[1] == '=') {
                        ThisApply->TruthStates[SDIR_EQUAL] = TRUE;
                    } else {
                        ThisApply->TruthStates[SDIR_EQUAL] = FALSE;
                    }
                } else if (Operator[0] == '!' && Operator[1] == '=') {
                    ThisApply->CompareFn = SdirOptions[i].CompareFn;
                    ThisApply->TruthStates[SDIR_LESS_THAN] = TRUE;
                    ThisApply->TruthStates[SDIR_GREATER_THAN] = TRUE;
                    ThisApply->TruthStates[SDIR_EQUAL] = FALSE;
                } else if (Operator[0] == '=') {
                    ThisApply->CompareFn = SdirOptions[i].CompareFn;
                    ThisApply->TruthStates[SDIR_LESS_THAN] = FALSE;
                    ThisApply->TruthStates[SDIR_GREATER_THAN] = FALSE;
                    ThisApply->TruthStates[SDIR_EQUAL] = TRUE;
                } else if (Operator[0] == '&') {
                    ThisApply->CompareFn = SdirOptions[i].BitwiseCompareFn;
                    ThisApply->TruthStates[SDIR_EQUAL] = TRUE;
                    ThisApply->TruthStates[SDIR_NOT_EQUAL] = FALSE;
                } else if (Operator[0] == '!' && Operator[1] == '&') {
                    ThisApply->CompareFn = SdirOptions[i].BitwiseCompareFn;
                    ThisApply->TruthStates[SDIR_EQUAL] = FALSE;
                    ThisApply->TruthStates[SDIR_NOT_EQUAL] = TRUE;
                } else {
                    SdirWriteString(_T("Operator error in: "));
                    goto error_return;
                }

                //
                //  Maybe the operator specified was valid but not supported
                //  by this type.
                //

                if (ThisApply->CompareFn == NULL) {
                    SdirWriteString(_T("Operator not supported in: "));
                    goto error_return;
                }

                //
                //  If we fail to capture this, ignore it and move on to the
                //  next entry, clobbering this structure again.
                //

                if (!SdirOptions[i].GenerateFromStringFn(&ThisApply->CompareEntry, Criteria)) {
                    SdirWriteString(_T("Could not parse evalutation criteria in: "));
                    goto error_return;
                }

                //
                //  Mark anything we're filtering by for collection
                //

                OptEnabled = (PUCHAR)Opts + SdirOptions[i].EnableOffset;
                *OptEnabled |= SDIR_OPT_COLLECT;

                ThisApply->Attribute = (WORD)SdirAttributeFromString(Attribute);

                ThisApply++;
                break;
            }
        }

next_element:

        if (Next) {
            This = Next + 1;
        } else {
            This = NULL;
        }
    }

    HeapFree(GetProcessHeap(), 0, SdirAttributeCombinedApplyString);
    return TRUE;

error_return:

    SdirWriteString(SingleElement);
    SdirWriteString(_T("\n"));

    HeapFree(GetProcessHeap(), 0, SdirAttributeCombinedApplyString);
    return FALSE;
}


BOOL
SdirApplyAttribute(
    __in PSDIR_DIRENT DirEnt,
    __out WORD *Attribute
    )
{
    PSDIR_ATTRIBUTE_APPLY ThisApply;
    WORD ThisAttribute = 0;

    if (SdirAttributeApply == NULL) {
        return FALSE;
    }
    
    ThisApply = SdirAttributeApply;

    while (ThisApply->CompareFn) {

        if (ThisApply->TruthStates[ThisApply->CompareFn(DirEnt, &ThisApply->CompareEntry)]) {
            ThisAttribute |= ThisApply->Attribute;
            if ((ThisAttribute & SDIR_ATTRIBUTE_CONTINUE) == 0) {

                if (ThisAttribute & SDIR_ATTRIBUTE_INVERT) {
                    ThisAttribute = (WORD)((ThisAttribute & 0xFF00) + ((ThisAttribute & 0x000F) << 4) + ((ThisAttribute & 0x00F0) >> 4));
                }

                *Attribute = ThisAttribute;
                return TRUE;
            }

            ThisAttribute = ThisAttribute & ~(SDIR_ATTRIBUTE_CONTINUE);
        }

        ThisApply++;
    }

    //
    //  We do let the user explicitly request black on black, but
    //  if we ended the search due to unbounded continues, return
    //  what we have.
    //

    if (ThisAttribute & SDIR_ATTRIBUTE_TERMINATE_MASK) {

        *Attribute = ThisAttribute;
        return TRUE;
    }
    
    return FALSE;
}

// vim:sw=4:ts=4:et:
