/*
 * INIT.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module implements initialization support including argument parsing
 * and initializing default options.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

BOOL WINAPI
SdirRestoreColor (
    __in DWORD dwHandlerType
    )
{
    Opts->Cancelled = TRUE;
    return TRUE;
}

BOOL
SdirAppInitialize()
{
    CONSOLE_SCREEN_BUFFER_INFO ScreenInfo;
    HANDLE hConsoleOutput = GetStdHandle(STD_OUTPUT_HANDLE);
    HANDLE hKernel32 = GetModuleHandle(_T("KERNEL32"));
    DISABLE_WOW_REDIRECT_FN DisableWow = (DISABLE_WOW_REDIRECT_FN)GetProcAddress(hKernel32, "Wow64DisableWow64FsRedirection");

    Opts = HeapAlloc(GetProcessHeap(), 0, sizeof(SDIR_OPTS));
    if (Opts == NULL) {
        return FALSE;
    }
    ZeroMemory(Opts, sizeof(SDIR_OPTS));

    Summary = HeapAlloc(GetProcessHeap(), 0, sizeof(SDIR_SUMMARY));
    if (Summary == NULL) {
        return FALSE;
    }
    ZeroMemory(Summary, sizeof(SDIR_SUMMARY));

    //
    //  For simplicity, initialize this now.  On failure we restore to
    //  this value.  Hopefully we'll find the correct value before any
    //  failure can occur.
    //

    Opts->PreviousAttributes = SDIR_DEFAULT_COLOR;

    //
    //  Try to determine the console width so we can size columns.  We
    //  try to use the window size, not the buffer size.
    //  If we're not talking to a console, this might fail, and we have
    //  to fall back to some kind of default.
    //

    if (!GetConsoleScreenBufferInfo(hConsoleOutput, &ScreenInfo)) {
        if (GetLastError() == ERROR_INVALID_HANDLE) {

            Opts->OutputIsConsole = FALSE;
            Opts->NewLine = _T("\r\n");
            if (Opts->ConsoleWidth == 0) {
                Opts->ConsoleWidth = 100;
            }
            Opts->ConsoleBufferWidth = Opts->ConsoleWidth;
            Opts->EnablePause = FALSE;
            Opts->PreviousAttributes = SDIR_DEFAULT_COLOR;

        } else {

            SdirDisplayError(GetLastError(), _T("GetConsoleScreenBufferInfo"));
            return FALSE;
        }

    } else {

        Opts->OutputIsConsole = TRUE;
        Opts->NewLine = _T("\n");
        Opts->EnablePause = TRUE;

        Opts->ConsoleBufferWidth = 
        Opts->ConsoleWidth = ScreenInfo.dwSize.X;
        
        if (Opts->ConsoleWidth > (DWORD)(ScreenInfo.srWindow.Right - ScreenInfo.srWindow.Left + 1)) {
            Opts->ConsoleWidth = (DWORD)(ScreenInfo.srWindow.Right - ScreenInfo.srWindow.Left + 1);
        }
    
        Opts->ConsoleHeight = ScreenInfo.dwSize.Y;
        if (Opts->ConsoleHeight > (DWORD)(ScreenInfo.srWindow.Bottom - ScreenInfo.srWindow.Top + 1)) {
            Opts->ConsoleHeight = (DWORD)(ScreenInfo.srWindow.Bottom - ScreenInfo.srWindow.Top + 1);
        }

        //
        //  This is another kludge.  To make progress we need to be able
        //  to display one line of output between "press any key" prompts,
        //  so we need to assume the console can hold two lines.  If we
        //  don't make this assumption, we'll never go forward.
        //

        if (Opts->ConsoleHeight < 2) {
            Opts->ConsoleHeight = 2;
        }

        Opts->PreviousAttributes = ScreenInfo.wAttributes;
    }

    if (Opts->ConsoleWidth > SDIR_MAX_WIDTH) {
        Opts->ConsoleWidth = SDIR_MAX_WIDTH;
    }

    //
    //  When running on WOW64, we don't want file system redirection,
    //  because we want people to be able to enumerate those paths
    //

    if (DisableWow) {
        PVOID DontCare;
        DisableWow(&DontCare);
    }

    //
    //  Try to locate FindFirstStream functions if they exist
    //

    Opts->FindFirstStreamW = (FIND_FIRST_STREAM_FN)GetProcAddress(hKernel32, "FindFirstStreamW");
    Opts->FindNextStreamW = (FIND_NEXT_STREAM_FN)GetProcAddress(hKernel32, "FindNextStreamW");
#ifdef UNICODE
    Opts->GetCompressedFileSize = (GET_COMPRESSED_FILE_SIZE_FN)GetProcAddress(hKernel32, "GetCompressedFileSizeW");
    Opts->GetDiskFreeSpaceEx = (GET_DISK_FREE_SPACE_EX_FN)GetProcAddress(hKernel32, "GetDiskFreeSpaceExW");
#else
    Opts->GetCompressedFileSize = (GET_COMPRESSED_FILE_SIZE_FN)GetProcAddress(hKernel32, "GetCompressedFileSizeA");
    Opts->GetDiskFreeSpaceEx = (GET_DISK_FREE_SPACE_EX_FN)GetProcAddress(hKernel32, "GetDiskFreeSpaceExA");
#endif
    Opts->GetFileInformationByHandleEx = (GET_FILE_INFORMATION_BY_HANDLE_EX_FN)GetProcAddress(hKernel32, "GetFileInformationByHandleEx");

    //
    //  Grab the version of the running OS so we can highlight binaries that
    //  need a newer one
    //

    Opts->OsVersion = GetVersion();

    //
    //  Look for Ctrl+C and restore color if we get one.
    //  Not much we can do on failure.
    //

    SetConsoleCtrlHandler(SdirRestoreColor, TRUE);

    return TRUE;
}


BOOL
SdirOptInitialize()
{
    ULONG i, j;

    //
    //  Calculate the amount of metadata in each column
    //

    Opts->MetadataWidth = 1;  // Column seperator

    //
    //  If no sorting algorithm was used, use the default that
    //  we prepopulated.
    //

    if (Opts->CurrentSort == 0) {
        Opts->CurrentSort++;
    }

    for (i = 0; i < SdirGetNumSdirOptions(); i++) {

        PUCHAR OptEnabled = (PUCHAR)Opts + SdirOptions[i].EnableOffset;

        //
        //  If we're displaying or sorting, we need the data to use
        //

        if ((*OptEnabled) & SDIR_OPT_DISPLAY) {

            *OptEnabled |= SDIR_OPT_COLLECT;
        }

        for (j = 0; j < Opts->CurrentSort; j++) {

            if (SdirOptions[i].CompareFn == Opts->Sort[j].CompareFn) {
                *OptEnabled |= SDIR_OPT_COLLECT;
            }
        }

        //
        //  If we're displaying, we need space to display in
        //  

        if (((*OptEnabled) & SDIR_OPT_DISPLAY) &&
               SdirOptions[i].WidthFn) {

            Opts->MetadataWidth += SdirOptions[i].WidthFn(NULL, 0, NULL);
        }
    }


    return TRUE;
}


BOOL
SdirParseOpt (
    __in LPTSTR Opt
    )
{
    BOOL OptParsed = FALSE;
    DWORD i, j;

    if (Opt[0] == 'd') {
        for (i = 0; i < SdirGetNumSdirOptions(); i++) {
            if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0 &&
                SdirOptions[i].Default & SDIR_OPT_ALLOW_DISPLAY) {

                PUCHAR OptEnabled = (PUCHAR)Opts + SdirOptions[i].EnableOffset;

                *OptEnabled |= SDIR_OPT_DISPLAY;
                OptParsed = TRUE;
                break;
            }
        }
    } else if (Opt[0] == 'h') {
        for (i = 0; i < SdirGetNumSdirOptions(); i++) {
            if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0 &&
                SdirOptions[i].Default & SDIR_OPT_ALLOW_DISPLAY) {

                PUCHAR OptEnabled = (PUCHAR)Opts + SdirOptions[i].EnableOffset;

                *OptEnabled &= ~(SDIR_OPT_DISPLAY);
                OptParsed = TRUE;
                break;
            }
        }
    } else if (Opt[0] == 's' || Opt[0] == 'i') {

        //
        //  If we don't have space for it, don't even try
        //

        if (Opts->CurrentSort < sizeof(Opts->Sort)/sizeof(Opts->Sort[0])) {

            //
            //  Find which option it is
            //

            for (i = 0; i < SdirGetNumSdirOptions(); i++) {
                if (_tcsicmp(&Opt[1], SdirOptions[i].Switch) == 0 &&
                    SdirOptions[i].Default & SDIR_OPT_ALLOW_SORT) {

                    //
                    //  See if it's been specified before and silently discard
                    //  it.
                    //

                    for (j = 0; j < Opts->CurrentSort; j++) {
                        if (Opts->Sort[j].CompareFn == SdirOptions[i].CompareFn) {
                            OptParsed = TRUE;
                            break;
                        }
                    }

                    if (OptParsed) {
                        break;
                    }

                    Opts->Sort[Opts->CurrentSort].CompareFn = SdirOptions[i].CompareFn;

                    if (Opt[0] == 's') {
                        Opts->Sort[Opts->CurrentSort].CompareBreakCondition = SDIR_GREATER_THAN;
                        Opts->Sort[Opts->CurrentSort].CompareInverseCondition = SDIR_LESS_THAN;
                    } else {
                        Opts->Sort[Opts->CurrentSort].CompareBreakCondition = SDIR_LESS_THAN;
                        Opts->Sort[Opts->CurrentSort].CompareInverseCondition = SDIR_GREATER_THAN;
                    }
                    Opts->CurrentSort++;
                    OptParsed = TRUE;

                    break;
                }
            }

        } else {
        }
    } else if (Opt[0] == 'p') {
        if (Opt[1] == 'n') {
            Opts->EnablePause = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            if (Opts->OutputIsConsole) {
                Opts->EnablePause = TRUE;
            }
            OptParsed = TRUE;
        }
    } else if (Opt[0] == 't') {
        if (Opt[1] == 'n') {
            Opts->EnableNameTruncation = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            Opts->EnableNameTruncation = TRUE;
            OptParsed = TRUE;
        }
#ifdef UNICODE
    } else if (Opt[0] == 'u') {
        if (Opt[1] == 'n') {
            Opts->OutputIsUnicode = FALSE;
            OptParsed = TRUE;
        } else if (Opt[1] == '\0') {
            Opts->OutputIsUnicode = TRUE;
            OptParsed = TRUE;
        }
#endif
    } else if (Opt[0] == 'c') {
        if (Opt[1] == 'w') {
            Opts->ConsoleWidth = SdirStringToNum32(&Opt[2], NULL);
            OptParsed = TRUE;
        }
    } else if (Opt[0] == 'r') {

        LPTSTR FinalPart;
        DWORD SizeCopied;
        LPTSTR SubDirPath = &Opt[1];

        OptParsed = TRUE;
        if (SubDirPath[0] == '\0') {
            SubDirPath = _T(".");
        }

        SizeCopied = GetFullPathName(SubDirPath,  sizeof(Opts->SubDirWalk)/sizeof(Opts->SubDirWalk[0]), Opts->SubDirWalk, &FinalPart);
        if (SizeCopied == 0 || SizeCopied >= sizeof(Opts->SubDirWalk)/sizeof(Opts->SubDirWalk[0])) {
            SdirDisplayError(GetLastError(), _T("GetFullPathName"));
            return OptParsed;
        }
    
        Opts->SubDirWalk[sizeof(Opts->SubDirWalk)/sizeof(Opts->SubDirWalk[0]) - 1] = '\0';
    }

    return OptParsed;
}

BOOL
SdirParseArgs (
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    ULONG  CurrentArg;
    BOOL   OptParsed;
    PUCHAR OptEnabled;

    TCHAR  EnvOpts[200];
    LPTSTR Opt;
    LPTSTR TokContext = NULL;
    DWORD  i;

    //
    //  Default to name sorting.  If something else is specified
    //  we clobber this entry.
    //

    Opts->CurrentSort = 0;
    Opts->Sort[0].CompareFn = SdirCompareFileName;
    Opts->Sort[0].CompareBreakCondition = SDIR_GREATER_THAN;
    Opts->Sort[0].CompareInverseCondition = SDIR_LESS_THAN;

    Opts->EnableNameTruncation = TRUE;

    Opts->SubDirWalk[0] = '\0';

    //
    //  Default to ANSI files
    //

    Opts->OutputIsUnicode = FALSE;

    for (i = 0; i < SdirGetNumSdirOptions(); i++) {
        OptEnabled = (PUCHAR)Opts + SdirOptions[i].EnableOffset;
        
        *OptEnabled = SdirOptions[i].Default;
    }

    if (GetEnvironmentVariable(_T("SDIR_OPTS"), EnvOpts, sizeof(EnvOpts)/sizeof(EnvOpts[0]))) {
        EnvOpts[sizeof(EnvOpts)/sizeof(EnvOpts[0]) - 1] = '\0';
        Opt = _tcstok_s(EnvOpts, _T(" "), &TokContext);

        while (Opt) {

            if (Opt[0] == '-' || Opt[0] == '/') {

                Opt++;

                OptParsed = FALSE;

                OptParsed = SdirParseOpt(Opt);
    
                if (!OptParsed) {
                    SdirWriteString(_T("Unknown environment option: "));
                    SdirWriteString(Opt);
                    SdirWriteString(_T("\n"));
                    SdirUsage();
                    return FALSE;
                }
            }

            Opt = _tcstok_s(NULL, _T(" "), &TokContext);
        }
    }

    for (CurrentArg = 1; CurrentArg < (ULONG)argc; CurrentArg++)
    {
        if (argv[CurrentArg][0] == '-' || argv[CurrentArg][0] == '/') {
            LPTSTR Opt = &argv[CurrentArg][1];

            OptParsed = FALSE;

            OptParsed = SdirParseOpt(Opt);

            if (_tcsicmp(Opt, _T("help")) == 0) {
                SdirUsage();
                return FALSE;
            } else if (_tcsicmp(Opt, _T("?")) == 0) {
                SdirUsage();
                return FALSE;
            }

            if (!OptParsed) {
                SdirWriteString(_T("Unknown argument: "));
                SdirWriteString(argv[CurrentArg]);
                SdirWriteString(_T("\n"));
                SdirUsage();
                return FALSE;
            }
        }
    }
    return TRUE;
}

BOOL
SdirInit(
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    if (!SdirAppInitialize()) {
        return FALSE;
    }

    if (!SdirParseArgs(argc, argv)) {
        return FALSE;
    }

    if (!SdirOptInitialize()) {
        return FALSE;
    }

    if (!SdirParseAttributeApplyString()) {
        return FALSE;
    }

    return TRUE;
}

// vim:sw=4:ts=4:et:
