/*
 * SDIR.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module implements the core logic of displaying directories.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

const SDIR_ATTRPAIR AttrPairs[] = {
    {FILE_ATTRIBUTE_ARCHIVE,           'A'},
    {FILE_ATTRIBUTE_READONLY,          'R'},
    {FILE_ATTRIBUTE_HIDDEN,            'H'},
    {FILE_ATTRIBUTE_SYSTEM,            'S'},
    {FILE_ATTRIBUTE_DIRECTORY,         'D'},
    {FILE_ATTRIBUTE_COMPRESSED,        'C'},
    {FILE_ATTRIBUTE_ENCRYPTED,         'E'},
    {FILE_ATTRIBUTE_OFFLINE,           'O'},
    {FILE_ATTRIBUTE_REPARSE_POINT,     'r'},
    {FILE_ATTRIBUTE_SPARSE_FILE,       's'},
    {FILE_ATTRIBUTE_INTEGRITY_STREAM,  'I'},
    };

DWORD
SdirGetNumAttrPairs()
{
    return sizeof(AttrPairs)/sizeof(AttrPairs[0]);
}


DWORD SdirAllocatedDirents     = 1000;
PSDIR_DIRENT SdirDirCollection = NULL;
PSDIR_DIRENT * SdirDirSorted   = NULL;
ULONG SdirDirCollectionCurrent = 0;
ULONG SdirDirCollectionLongest = 0;
ULONG SdirDirCollectionTotalNameLength = 0;

PSDIR_OPTS Opts = NULL;
PSDIR_SUMMARY Summary = NULL;


BOOL
SdirDisplayCollection();

BOOL
SdirAddToCollection (
    __in PWIN32_FIND_DATA FindData,
    __in LPTSTR FullPath
    ) 
{
    PSDIR_DIRENT CurrentEntry;
    DWORD i, j;
    PUCHAR OptEnabled;
    DWORD CompareResult;

    if (SdirDirCollectionCurrent >= SdirAllocatedDirents) {
        if (SdirDirCollectionCurrent < UINT_MAX) {
            SdirDirCollectionCurrent++;
        }
        return FALSE;
    }

    CurrentEntry = &SdirDirCollection[SdirDirCollectionCurrent];

    SdirDirCollectionCurrent++;

    //
    //  Copy over the data from Win32's FindFirstFile into our own structure.
    //

    for (i = 0; i < SdirGetNumSdirOptions(); i++) {

        OptEnabled = (PUCHAR)Opts + SdirOptions[i].EnableOffset;

        //
        //  If we're displaying or sorting, we need the data to use
        //

        if (((*OptEnabled) & SDIR_OPT_COLLECT) &&
               SdirOptions[i].CollectFn) {

            SdirOptions[i].CollectFn(CurrentEntry, FindData, FullPath);
        }
    }

    //
    //  Determine the color to display each entry from extensions and attributes.
    //  If we're asked to hide, just walk away from the entry we created and
    //  continue.
    //
    
    if (!SdirApplyAttribute(CurrentEntry, &CurrentEntry->RenderAttributes)) {
        CurrentEntry->RenderAttributes = SDIR_DEFAULT_COLOR;
    } else if (CurrentEntry->RenderAttributes & SDIR_ATTRIBUTE_HIDE) {
        SdirDirCollectionCurrent--;
        return TRUE;
    }

    if (Opts->EnableSummary & SDIR_OPT_COLLECT) {
        SdirCollectSummary(CurrentEntry);
    }

    //
    //  Now that our internal entry is fully poulated, insert it into the correct sorted position.
    //  As an optimization, check if we just need to insert at the end (for file name sort on
    //  NTFS, this is the common case.)
    //

    if (SdirDirCollectionCurrent > 1 &&
        Opts->Sort[0].CompareFn(SdirDirSorted[SdirDirCollectionCurrent - 2], CurrentEntry) == Opts->Sort[0].CompareInverseCondition) {

        SdirDirSorted[SdirDirCollectionCurrent - 1] = CurrentEntry;
        return TRUE;
    }

    //
    //  MSFIX This algorithm seems really stupid.  Since we know the sort
    //  list is sorted, shouldn't we binary search?
    //

    for (i = 0; i < SdirDirCollectionCurrent - 1; i++) {

        for (j = 0; j < Opts->CurrentSort; j++) {
            CompareResult = Opts->Sort[j].CompareFn(SdirDirSorted[i], CurrentEntry);

            if (CompareResult == Opts->Sort[j].CompareBreakCondition ||
                CompareResult == Opts->Sort[j].CompareInverseCondition) {
                break;
            }
        }

        if (CompareResult == Opts->Sort[j].CompareBreakCondition) {
            for (j = SdirDirCollectionCurrent - 1; j > i; j--) {
                SdirDirSorted[j] = SdirDirSorted[j - 1];
            }
            SdirDirSorted[i] = CurrentEntry;
            return TRUE;
        }
    }
    SdirDirSorted[SdirDirCollectionCurrent - 1] = CurrentEntry;
    return TRUE;
}

BOOL
SdirEnumeratePath (
    __in LPCTSTR FindStr
    )
{
    HANDLE hFind;
    WIN32_FIND_DATA FindData;
    LPTSTR FinalPart;
    DWORD SizeCopied;
    TCHAR FullPath[MAX_PATH];

    //
    //  At this point we should have a directory and an enumeration criteria.
    //  Copy the directory name and clobber the enumeration criteria.
    //
    
    SizeCopied = GetFullPathName(FindStr, sizeof(Opts->ParentName)/sizeof(Opts->ParentName[0]), Opts->ParentName, &FinalPart);
    if (SizeCopied == 0 || SizeCopied > sizeof(Opts->ParentName)/sizeof(Opts->ParentName[0])) {
        SdirDisplayError(GetLastError(), _T("GetFullPathName"));
        return FALSE;
    }
    Opts->ParentNameLength = SizeCopied;
    *FinalPart = '\0';

    if (Opts->EnableSummary & SDIR_OPT_COLLECT) {

        LARGE_INTEGER Junk;

        //
        //  If GetDiskFreeSpaceEx fails, fall back to GetDiskFreeSpace.  According
        //  to MSDN, this can sometimes be present and fail with not supported.
        //

        if (Opts->GetDiskFreeSpaceEx == NULL ||
            !Opts->GetDiskFreeSpaceEx(Opts->ParentName, &Junk, &Summary->VolumeSize, &Summary->FreeSize)) {

            ULONG BytesPerSector;
            ULONG SectorsPerCluster;
            ULONG FreeClusters;
            ULONG TotalClusters;

            GetDiskFreeSpace(Opts->ParentName, &SectorsPerCluster, &BytesPerSector, &FreeClusters, &TotalClusters);

#if SDIR_BIG_SIZE
            Summary->VolumeSize.QuadPart = SectorsPerCluster * BytesPerSector * TotalClusters;
            Summary->FreeSize.QuadPart = SectorsPerCluster * BytesPerSector * FreeClusters;
#else
            Summary->VolumeSize.LowPart = SectorsPerCluster * BytesPerSector * TotalClusters;
            Summary->FreeSize.LowPart = SectorsPerCluster * BytesPerSector * FreeClusters;
#endif
        }
    }

    //
    //  We loop enumerating all the files.  Hopefully for common directories
    //  we'll allocate a big enough buffer in the first case and we can then
    //  just populate that buffer and display it.  If the directory is large
    //  enough though, we count the number of entries, loop back, and allocate
    //  a large enough buffer to hold the result, then enumerate it again.
    //  Because we sort the output, we must keep the entire set in memory to
    //  be able to meaningfully process it.
    //

    do {

        //
        //  If this is a subsequent pass and we didn't allocate a big enough
        //  buffer, reallocate our buffers.  Add an extra 100 just in case
        //  we're still adding files in real time.
        //

        if (SdirDirCollectionCurrent >= SdirAllocatedDirents) {
            SdirAllocatedDirents = SdirDirCollectionCurrent;
            if (SdirAllocatedDirents < UINT_MAX - 100) {
                SdirAllocatedDirents += 100;
            }
        }

        SdirDirCollectionCurrent = 0;

        if (SdirDirCollection != NULL) {
            HeapFree(GetProcessHeap(), 0, SdirDirCollection);
        }

        if (SdirDirSorted != NULL) {
            HeapFree(GetProcessHeap(), 0, SdirDirSorted);
        }

        SdirDirCollection = HeapAlloc(GetProcessHeap(), 0, SdirAllocatedDirents * sizeof(SDIR_DIRENT));
        if (SdirDirCollection == NULL) {
            SdirDisplayError(GetLastError(), _T("HeapAlloc"));
            return FALSE;
        }

        SdirDirSorted = HeapAlloc(GetProcessHeap(), 0, SdirAllocatedDirents * sizeof(PSDIR_DIRENT));
        if (SdirDirSorted == NULL) {
            SdirDisplayError(GetLastError(), _T("HeapAlloc"));
            return FALSE;
        }

        
        hFind = FindFirstFile(FindStr, &FindData);
        
        //
        //  If we can't find enumerate, display the error except when we're recursive
        //  and the error is we found no files in this particular directory.
        //

        if (hFind == NULL || hFind == INVALID_HANDLE_VALUE) {

            if (Opts->SubDirWalk[0] == '\0' || GetLastError() != ERROR_FILE_NOT_FOUND) {
                SdirDisplayError(GetLastError(), NULL);
            }
            return FALSE;
        }
        
        do {

            _stprintf_s(FullPath, MAX_PATH, _T("%s%s"), Opts->ParentName, FindData.cFileName);
            FullPath[MAX_PATH - 1] = '\0';

#ifdef UNICODE
            if (Opts->FindFirstStreamW && (Opts->EnableNamedStreams & SDIR_OPT_DISPLAY)) {
                HANDLE hStreamFind;
                HANDLE hFile;
                SDIR_WIN32_FIND_STREAM_DATA FindStreamData;
                WIN32_FIND_DATA BogusFindData;
                BY_HANDLE_FILE_INFORMATION FileInfo;

                //
                //  Display the default stream
                //

                SdirAddToCollection(&FindData, FullPath);

                //
                //  Look for any named streams
                //

                hStreamFind = Opts->FindFirstStreamW(FullPath, 0, &FindStreamData, 0);
                if (hStreamFind != INVALID_HANDLE_VALUE) {

                    do {

                        //
                        //  For the default stream, just report the information we found
                        //  for the file.  For anything else, query all the equivalent
                        //  stream information
                        //

                        if (_tcscmp(FindStreamData.cStreamName, L"::$DATA") != 0) {

                            DWORD StreamLength = _tcslen(FindStreamData.cStreamName);

                            if (StreamLength > 6 && _tcscmp(FindStreamData.cStreamName + StreamLength - 6, L":$DATA") == 0) {
                                FindStreamData.cStreamName[StreamLength - 6] = '\0';
                            }

                            _stprintf_s(FullPath, MAX_PATH, _T("%s%s%s"), Opts->ParentName, FindData.cFileName, FindStreamData.cStreamName);

                            //
                            //  Assume file state is stream state
                            //

                            memcpy(&BogusFindData, &FindData, sizeof(FindData));

                            //
                            //  Populate stream name
                            //

                            _stprintf_s(BogusFindData.cFileName, MAX_PATH, _T("%s%s"), FindData.cFileName, FindStreamData.cStreamName);

                            //
                            //  Populate stream information
                            //

                            hFile = CreateFile(FullPath,
                                               FILE_READ_ATTRIBUTES,
                                               FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
                                               NULL,
                                               OPEN_EXISTING,
                                               FILE_FLAG_BACKUP_SEMANTICS,
                                               NULL);

                            if (hFile != INVALID_HANDLE_VALUE) {

                                GetFileInformationByHandle(hFile, &FileInfo);

                                BogusFindData.dwFileAttributes = FileInfo.dwFileAttributes;
                                BogusFindData.ftCreationTime = FileInfo.ftCreationTime;
                                BogusFindData.ftLastAccessTime = FileInfo.ftLastAccessTime;
                                BogusFindData.ftLastWriteTime = FileInfo.ftLastWriteTime;
                                BogusFindData.nFileSizeHigh = FileInfo.nFileSizeHigh;
                                BogusFindData.nFileSizeLow  = FileInfo.nFileSizeLow;

                                CloseHandle(hFile);
                            }
                            SdirAddToCollection(&BogusFindData, FullPath);
                        }
                    } while (Opts->FindNextStreamW(hStreamFind, &FindStreamData));
                }

                FindClose(hStreamFind);

            } else {
#endif
                SdirAddToCollection(&FindData, FullPath);
#ifdef UNICODE
            }
#endif
        } while (FindNextFile(hFind, &FindData) && !Opts->Cancelled);
        
        FindClose(hFind);

        //
        //  If we've encountered more than we've allocated, go back
        //  and reallocate.
        //

    } while (SdirDirCollectionCurrent >= SdirAllocatedDirents);

    return TRUE;
}


#define SDIR_LINE_ELEMENT_HORIZ      0
#define SDIR_LINE_ELEMENT_TOP_T      1
#define SDIR_LINE_ELEMENT_BOTTOM_T   2
#define SDIR_LINE_ELEMENT_VERT       3

#ifdef UNICODE
TCHAR SdirLineElementsRich[] = {0x2500, 0x252c, 0x2534, 0x2502};
#endif
TCHAR SdirLineElementsText[] = {'-', '+', '+', '|'};

DWORD SdirDisplayedRows = 0;

BOOL
SdirRowDisplayed()
{
    if (Opts->EnablePause) {
        SdirDisplayedRows++;

        if (SdirDisplayedRows >= Opts->ConsoleHeight - 1) {

            if (!SdirPressAnyKey()) {
                return FALSE;
            }
            SdirDisplayedRows = 0;
        }
    }
    return TRUE;
}


BOOL
SdirDisplayCollection()
{
    PSDIR_DIRENT CurrentEntry;
    DWORD Index, Ext;
    WORD Attributes;
    DWORD Columns;
    DWORD ColumnWidth;
    DWORD ActiveColumn = 0;
    SDIR_FMTCHAR Line[SDIR_MAX_WIDTH];
    DWORD CurrentChar = 0;
    DWORD BufferRows;
    DWORD LongestDisplayedFileName = SdirDirCollectionLongest;
    LPTSTR LineElements = SdirLineElementsText;
    PUCHAR OptEnabled;

#ifdef UNICODE
    if (Opts->OutputIsConsole) {
        LineElements = SdirLineElementsRich;
    }
#endif

    //
    //  If we're allowed to shorten names to make the display more
    //  legible, we won't allow a longest name greater than twice
    //  average size.  Since we need to insert three characters to
    //  indicate we butchered a name, don't do this unless we have
    //  a meaningful length to start with (currently 10.)
    //

    if (Opts->EnableNameTruncation) {
        ULONG AverageNameLength;

        AverageNameLength = SdirDirCollectionTotalNameLength / SdirDirCollectionCurrent;
        if (LongestDisplayedFileName > 2 * AverageNameLength && AverageNameLength > 10) {
            LongestDisplayedFileName = 2 * AverageNameLength;
        }
    }

    if (Opts->EnableFileName & SDIR_OPT_DISPLAY) {
        Columns = Opts->ConsoleWidth / (LongestDisplayedFileName + Opts->MetadataWidth);
    } else {
        Columns = Opts->ConsoleWidth / (Opts->MetadataWidth);
    }

    //
    //  If the output is too big to fit the display, force one column
    //  of sufficient size.  The output here will be scrambled, but it
    //  will indicate to the user what happened.
    //

    if (Columns > 0) {
        ColumnWidth = Opts->ConsoleWidth / Columns;
        LongestDisplayedFileName = ColumnWidth - Opts->MetadataWidth;
    } else {
        Columns = 1;
        ColumnWidth = Opts->MetadataWidth;
        if (Opts->EnableFileName & SDIR_OPT_DISPLAY) {
            ColumnWidth += LongestDisplayedFileName;
        }
    }

    //
    //  This really shouldn't happen, even in the worst case of a MAX_PATH name
    //  with all metadata options enabled, but we'll be paranoid.
    //

    if (Columns * ColumnWidth >= SDIR_MAX_WIDTH) {
        SdirWriteString(_T("Too much data for a single line!"));
        return FALSE;
    }

    BufferRows = (SdirDirCollectionCurrent + Columns - 1) / Columns;

    //
    //  Draw the top grid line.
    //

    for (Index = 0; Index < ColumnWidth * Columns; Index++) {
        if (Index % ColumnWidth == ColumnWidth - 1 && Index < (ColumnWidth*Columns-1)) {
            Line[Index].Char = LineElements[SDIR_LINE_ELEMENT_TOP_T];
            Line[Index].Attr = SDIR_GRID_COLOR;
        } else {
            Line[Index].Char = LineElements[SDIR_LINE_ELEMENT_HORIZ];
            Line[Index].Attr = SDIR_GRID_COLOR;
        }
    }

    SdirWrite(Line, Index);

    if (ColumnWidth * Columns != Opts->ConsoleBufferWidth || !Opts->OutputIsConsole) {
        SdirWriteString(_T("\n"));
    }

    if (!SdirRowDisplayed()) {
        return FALSE;
    }

    //
    //  Enumerate through the entries.
    //

    for (Index = 0; Index < BufferRows * Columns && !Opts->Cancelled; Index++) {

        //
        //  Because we're sorting down columns first, but rendering a row at a time,
        //  we need to do some matrix math to find which elements belong in which cells.
        //  Some cells in the bottom right might be empty.
        //

        Ext = ActiveColumn * BufferRows + Index / Columns;
        if (Ext < SdirDirCollectionCurrent) {
            CurrentEntry = SdirDirSorted[Ext];
        } else {
            CurrentEntry = NULL;
        }
        Attributes = SDIR_DEFAULT_COLOR;

        //
        //  Render the empty cell.  Or, if we have contents, render that too.
        //

        if (CurrentEntry == NULL) {
            SdirPasteStrAndPad(&Line[CurrentChar], NULL, Attributes, 0, ColumnWidth - 1);
            CurrentChar += ColumnWidth - 1;
        } else {

            Attributes = CurrentEntry->RenderAttributes;
    
            //
            //  Paste file name into buffer
            //
    
            if (Opts->EnableFileName & SDIR_OPT_DISPLAY) {
                if (CurrentEntry->FileNameLengthInChars > LongestDisplayedFileName) {
                    ULONG ExtractedLength = (LongestDisplayedFileName - 3) / 2;

                    SdirPasteStr(&Line[CurrentChar], CurrentEntry->FileName, Attributes, ExtractedLength);
                    CurrentChar += ExtractedLength;

                    SdirPasteStr(&Line[CurrentChar], _T("..."), Attributes, ExtractedLength);
                    CurrentChar += 3;

                    SdirPasteStrAndPad(&Line[CurrentChar],
                                       CurrentEntry->FileName + CurrentEntry->FileNameLengthInChars - ExtractedLength,
                                       Attributes,
                                       ExtractedLength,
                                       ColumnWidth - Opts->MetadataWidth - ExtractedLength - 3);

                    CurrentChar += ColumnWidth - Opts->MetadataWidth - ExtractedLength - 3;
                } else {
#pragma prefast(suppress: __WARNING_INCORRECT_VALIDATION, File name length is capped and used to calculate column width, which is capped.)
                    SdirPasteStrAndPad(&Line[CurrentChar],
                                       CurrentEntry->FileName,
                                       Attributes,
                                       CurrentEntry->FileNameLengthInChars,
                                       ColumnWidth - Opts->MetadataWidth);
                    CurrentChar += ColumnWidth - Opts->MetadataWidth;
                }
            }

            if (Opts->EnableShortName & SDIR_OPT_DISPLAY) {
                CurrentChar += SdirDisplayShortName(&Line[CurrentChar], Attributes, CurrentEntry);

                if (!(Opts->EnableFileName & SDIR_OPT_DISPLAY)) {

                    //
                    //  If file name is hidden, we may need to align things, because
                    //  our columns may contain padding (they're essentially justified.)
                    //

#pragma prefast(suppress: __WARNING_INCORRECT_VALIDATION, ColumnWidth is maximum file name size (MAX_PATH) plus metadata, padded up to terminal width, which is capped at the size of this buffer.)
                    SdirPasteStrAndPad(&Line[CurrentChar], NULL, 0, 0, ColumnWidth - Opts->MetadataWidth);
                    CurrentChar += ColumnWidth - Opts->MetadataWidth;
                }
            }

            //
            //  Paste any metadata options into the buffer.
            //

            for (Ext = 0; Ext < SdirGetNumSdirExec(); Ext++) {
                OptEnabled = (PUCHAR)Opts + SdirExec[Ext].EnableOffset;
                if ((*OptEnabled) & SDIR_OPT_DISPLAY) {
#pragma prefast(suppress: __WARNING_INCORRECT_VALIDATION, All of this adds up to MetadataWidth, which we sanity checked before)
                    CurrentChar += SdirExec[Ext].Function(&Line[CurrentChar], Attributes, CurrentEntry);
                }
            }
        }
    
        //
        //  We're starting a new column.  If it's the final one we might want a newline,
        //  otherwise we might want a gridline.
        //

        ActiveColumn++;
        if (ActiveColumn % Columns == 0) {

            for (Ext = 0; Opts->NewLine[Ext] != '\0'; Ext++) {
                Line[CurrentChar].Char = Opts->NewLine[Ext];
                Line[CurrentChar].Attr = SDIR_DEFAULT_COLOR;
                CurrentChar++;
            }
            SdirWrite(Line, CurrentChar);
            CurrentChar = 0;
            ActiveColumn = 0;
            if (!SdirRowDisplayed()) {
                return FALSE;
            }
        } else {
            Line[CurrentChar].Char = LineElements[SDIR_LINE_ELEMENT_VERT];
            Line[CurrentChar].Attr = SDIR_GRID_COLOR;
            CurrentChar++;
        }
    }

    //
    //  Render the bottom gridline.
    //

    for (Index = 0; Index < ColumnWidth * Columns; Index++) {
        if (Index % ColumnWidth == ColumnWidth - 1 && Index < (ColumnWidth*Columns-1)) {
            Line[Index].Char = LineElements[SDIR_LINE_ELEMENT_BOTTOM_T];
            Line[Index].Attr = SDIR_GRID_COLOR;
        } else {
            Line[Index].Char = LineElements[SDIR_LINE_ELEMENT_HORIZ];
            Line[Index].Attr = SDIR_GRID_COLOR;
        }
    }

    SdirWrite(Line, Index);
    CurrentChar = 0;

    //
    //  We want to restore the 'correct' color back when we return to CMD.
    //

    if (Opts->OutputIsConsole) {
        SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), SDIR_DEFAULT_COLOR);
    }

    if (ColumnWidth * Columns != Opts->ConsoleBufferWidth || !Opts->OutputIsConsole) {
        SdirWriteString(_T("\n"));
    }

    if (!SdirRowDisplayed()) {
        return FALSE;
    }

    return TRUE;
}

BOOL
SdirEnumerateAndDisplay (
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    TCHAR FindStr[MAX_PATH];
    LPTSTR FinalPart;
    DWORD SizeCopied;
    LPTSTR EnumPath;
    ULONG  CurrentArg;
    BOOLEAN EnumerateUserSpecified = FALSE;

    for (CurrentArg = 1; CurrentArg < (ULONG)argc; CurrentArg++) {

        if (argv[CurrentArg][0] != '-' && argv[CurrentArg][0] != '/') {

            HANDLE hDir;

            EnumPath = argv[CurrentArg];
            EnumerateUserSpecified = TRUE;

            //
            //  Start by assuming the input needs no processing
            //
    
            SizeCopied = GetFullPathName(EnumPath,  sizeof(FindStr)/sizeof(FindStr[0]), FindStr, &FinalPart);
            if (SizeCopied == 0 || SizeCopied >= sizeof(FindStr)/sizeof(FindStr[0])) {
                SdirDisplayError(GetLastError(), _T("GetFullPathName"));
                return FALSE;
            }
    
            FindStr[sizeof(FindStr)/sizeof(FindStr[0]) - 1] = '\0';
    
            hDir = CreateFile(FindStr, FILE_READ_ATTRIBUTES, FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, NULL);
    
            if (hDir != INVALID_HANDLE_VALUE) {
                BY_HANDLE_FILE_INFORMATION HandleInfo;
    
                GetFileInformationByHandle(hDir, &HandleInfo);
    
                if (HandleInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                    _stprintf_s(FindStr + SizeCopied, MAX_PATH - SizeCopied, _T("\\*"));
                }
    
                CloseHandle(hDir);
            }

            FindStr[MAX_PATH - 1] = '\0';

            if (!SdirEnumeratePath( FindStr )) {
                return FALSE;
            }

        }
    }

    if (!EnumerateUserSpecified) {
        if (!SdirEnumeratePath( _T("*") )) {
            return FALSE;
        }
    }


    if (SdirDirCollectionCurrent == 0) {
        SdirDisplayError(ERROR_FILE_NOT_FOUND, NULL);
        return FALSE;
    }

    if (!SdirDisplayCollection()) {
        return FALSE;
    }

    return TRUE;
}

BOOL
SdirEnumerateAndDisplaySubtree (
    __in TCHAR * TreeRoot,
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    TCHAR NextSubDir[MAX_PATH];
    HANDLE hFind;
    WIN32_FIND_DATA FindData;
    ULONG  CurrentArg;
    BOOLEAN EnumerateUserSpecified = FALSE;

    //
    //  If the user has specified search criteria, go ahead and collect
    //  those.  There may be multiple.
    //

    for (CurrentArg = 1; CurrentArg < (ULONG)argc; CurrentArg++) {

        if (argv[CurrentArg][0] != '-' && argv[CurrentArg][0] != '/') {

            _stprintf_s(NextSubDir, sizeof(NextSubDir)/sizeof(NextSubDir[0]), _T("%s\\%s"), TreeRoot, argv[CurrentArg]);

            if (NextSubDir[0] == '\0') {
                SdirWriteString(_T("Path exceeds MAX_PATH\n"));
                return FALSE;
            }

            EnumerateUserSpecified = TRUE;

            if (!SdirEnumeratePath( NextSubDir ) && GetLastError() != ERROR_FILE_NOT_FOUND) {
                return FALSE;
            }
        }
    }

    //
    //  If the user hasn't specified any search criteria, assume "*" and
    //  collect everything.
    //

    if (!EnumerateUserSpecified) {
        _stprintf_s(NextSubDir, sizeof(NextSubDir)/sizeof(NextSubDir[0]), _T("%s\\*"), TreeRoot);

        if (NextSubDir[0] == '\0') {
            SdirWriteString(_T("Path exceeds MAX_PATH\n"));
            return FALSE;
        }

        if (!SdirEnumeratePath( NextSubDir ) && GetLastError() != ERROR_FILE_NOT_FOUND) {
            return FALSE;
        }
    }

    //
    //  If we have something to display, display it.
    //

    if (SdirDirCollectionCurrent > 0) {

        SdirWriteStringWithAttribute(TreeRoot, SDIR_DIR_COLOR);
        SdirWriteString(_T("\n"));

        if (!SdirRowDisplayed()) {
            return FALSE;
        }

        if (!SdirDisplayCollection()) {
            return FALSE;
        }
    }

    SdirDirCollectionCurrent = 0;
    SdirDirCollectionLongest = 0;
    SdirDirCollectionTotalNameLength = 0;

    _stprintf_s(NextSubDir, sizeof(NextSubDir)/sizeof(NextSubDir[0]), _T("%s\\*"), TreeRoot);

    if (NextSubDir[0] == '\0') {
        SdirWriteString(_T("Path exceeds MAX_PATH\n"));
        return FALSE;
    }

    hFind = FindFirstFile(NextSubDir, &FindData);
    
    if (hFind == NULL || hFind == INVALID_HANDLE_VALUE) {
        SdirDisplayError(GetLastError(), NULL);
        return FALSE;
    }
    
    do {
        if (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY &&
            _tcscmp(FindData.cFileName, _T(".")) != 0 &&
            _tcscmp(FindData.cFileName, _T("..")) != 0) {

            _stprintf_s(NextSubDir, sizeof(NextSubDir)/sizeof(NextSubDir[0]), _T("%s\\%s"), TreeRoot, FindData.cFileName);

            if (NextSubDir[0] == '\0') {
                SdirWriteString(_T("Path exceeds MAX_PATH\n"));
                return FALSE;
            }

            if (!SdirEnumerateAndDisplaySubtree( NextSubDir, argc, argv )) {
                FindClose(hFind);
                return FALSE;
            }
            
        }
    } while (FindNextFile(hFind, &FindData) && !Opts->Cancelled);
    
    FindClose(hFind);

    return TRUE;
}


#ifdef UNICODE
#define ENTRYPOINT wmain
#else
#define ENTRYPOINT main
#endif

int __cdecl
ENTRYPOINT (
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    )
{
    if (!SdirInit(argc, argv)) {
        goto restore_and_exit;
    }

    if (Opts->SubDirWalk[0] != '\0') {
        if (!SdirEnumerateAndDisplaySubtree(Opts->SubDirWalk, argc, argv)) {
            goto restore_and_exit;
        }
    } else {
        if (!SdirEnumerateAndDisplay(argc, argv)) {
            goto restore_and_exit;
        }
    }

    if (Opts->EnableSummary & SDIR_OPT_DISPLAY) {

        SdirDisplaySummary(SDIR_DEFAULT_COLOR);
    }

restore_and_exit:

    if (Opts->OutputIsConsole) {

        SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), Opts->PreviousAttributes);
    }

    return 0;
}

// vim:sw=4:ts=4:et:
