/*
 * SDIR.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN 1
#endif

#include <windows.h>
#include <winioctl.h>
#ifndef MINICRT
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#endif
#include <tchar.h>

#ifdef MINICRT
#include <minicrt.h>
#endif


//
//  Compile time configuration
//

#define SDIR_MAX_WIDTH   500

#define SDIR_GRID_COLOR           (FOREGROUND_GREEN)
#define SDIR_FILE_SIZE_COLOR      (FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_INTENSITY)
#define SDIR_FILE_TIME_COLOR      (FOREGROUND_RED|FOREGROUND_GREEN)
#define SDIR_DEFAULT_COLOR        (FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE)
#define SDIR_DIR_COLOR            (FOREGROUND_RED|FOREGROUND_BLUE|FOREGROUND_INTENSITY)
#define SDIR_FUTURE_VERSION_COLOR (FOREGROUND_RED|FOREGROUND_INTENSITY)

//
//  Hacks for older compilers
//

#if (defined(_MSC_VER) && _MSC_VER >= 1200 && _MSC_VER < 1310)
#pragma comment(linker, "/OPT:NOWIN98")
#endif

#ifdef _ULONGLONG_
#define SDIR_BIG_SIZE 1
typedef ULONGLONG SDIR_FILESIZE;
#elif !defined(_MSC_VER) || _MSC_VER >= 900
#define SDIR_BIG_SIZE 1
typedef LONGLONG SDIR_FILESIZE;
#else
#pragma message("64 bit file size support not present")
typedef ULONG SDIR_FILESIZE;
#endif

#ifndef FILE_SHARE_DELETE
#define FILE_SHARE_DELETE 4
#endif

#ifndef FILE_ATTRIBUTE_SPARSE_FILE
#define FILE_ATTRIBUTE_SPARSE_FILE        (0x200)
#endif

#ifndef FILE_ATTRIBUTE_REPARSE_POINT
#define FILE_ATTRIBUTE_REPARSE_POINT      (0x400)
#endif

#ifndef FILE_ATTRIBUTE_COMPRESSED
#define FILE_ATTRIBUTE_COMPRESSED         (0x800)
#endif

#ifndef FILE_ATTRIBUTE_OFFLINE
#define FILE_ATTRIBUTE_OFFLINE           (0x1000)
#endif

#ifndef FILE_ATTRIBUTE_ENCRYPTED
#define FILE_ATTRIBUTE_ENCRYPTED         (0x4000)
#endif

#ifndef FILE_ATTRIBUTE_INTEGRITY_STREAM
#define FILE_ATTRIBUTE_INTEGRITY_STREAM  (0x8000)
#endif

#ifndef FSCTL_READ_FILE_USN_DATA
#define FSCTL_READ_FILE_USN_DATA         CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 58,  METHOD_NEITHER, FILE_ANY_ACCESS)
#endif

//
//  These were added in Vista but because they're not preprocessor
//  they're very hard to test for, so we test for something else
//

#ifndef FILE_ATTRIBUTE_VIRTUAL
typedef struct _FILE_STANDARD_INFO {
    LARGE_INTEGER AllocationSize;
    LARGE_INTEGER EndOfFile;
    DWORD NumberOfLinks;
    BOOLEAN DeletePending;
    BOOLEAN Directory;
} FILE_STANDARD_INFO, *PFILE_STANDARD_INFO;

#define FileStandardInfo    (0x000000001)
#endif

#ifndef _stprintf_s
#define _stprintf_s _sntprintf
#endif

#ifndef _tcstok_s
#define _tcstok_s(a,b,c) _tcstok(a,b)
#endif

#ifndef __in
#define __in
#define __out
#define __out_opt
#define __inout
#define __in_ecount(x)
#define __out_ecount(x)
#pragma warning(disable: 4068) // unknown pragma
#endif

typedef struct {

    DWORD          RecordLength;
    WORD           MajorVersion;
    WORD           MinorVersion;
    DWORDLONG      FileReferenceNumber;
    DWORDLONG      ParentFileReferenceNumber;
    LARGE_INTEGER  Usn;
    LARGE_INTEGER  TimeStamp;
    DWORD          Reason;
    DWORD          SourceInfo;
    DWORD          SecurityId;
    DWORD          FileAttributes;
    WORD           FileNameLength;
    WORD           FileNameOffset;
    WCHAR          FileName[1];

} SDIR_USN_RECORD, *PSDIR_USN_RECORD;

typedef struct _SDIR_WIN32_FIND_STREAM_DATA {
    LARGE_INTEGER StreamSize;
    WCHAR cStreamName[MAX_PATH + 36];
} SDIR_WIN32_FIND_STREAM_DATA, *PSDIR_WIN32_FIND_STREAM_DATA;

typedef struct _SDIR_ATTRPAIR {
    DWORD FileAttribute;
    TCHAR DisplayLetter;
} SDIR_ATTRPAIR, *PSDIR_ATTRPAIR;

extern const SDIR_ATTRPAIR AttrPairs[];

//
//  Run time configuration data
//

typedef struct _SDIR_FMTCHAR {
    TCHAR Char;
    WORD  Attr;
} SDIR_FMTCHAR, *PSDIR_FMTCHAR;

#define SDIR_ATTRIBUTE_INVERT           0x100
#define SDIR_ATTRIBUTE_HIDE             0x200
#define SDIR_ATTRIBUTE_CONTINUE         0x400
#define SDIR_ATTRIBUTE_COLOR_MASK       0xFF
#define SDIR_ATTRIBUTE_TERMINATE_MASK   (SDIR_ATTRIBUTE_COLOR_MASK|SDIR_ATTRIBUTE_HIDE)

typedef struct _SDIR_ATTRIBUTE_COLOR_STRING {
    LPTSTR String;
    WORD Attribute;
} SDIR_ATTRIBUTE_COLOR_STRING, *PSDIR_ATTRIBUTE_COLOR_STRING;

extern SDIR_ATTRIBUTE_COLOR_STRING ColorString[];

typedef struct _SDIR_DIRENT {
    LARGE_INTEGER FileId;
    LARGE_INTEGER AllocationSize;
    LARGE_INTEGER FileSize;
    LARGE_INTEGER CompressedFileSize;
    LARGE_INTEGER Usn;
    LARGE_INTEGER FileVersion;
    DWORD         FileVersionFlags;
    WORD          OsVersionHigh;
    WORD          OsVersionLow;
    SYSTEMTIME    CreateTime;
    SYSTEMTIME    WriteTime;
    DWORD         LinkCount;
    DWORD         StreamCount;
    DWORD         ReparseTag;
    DWORD         FileAttributes;
    WORD          RenderAttributes;
    DWORD         FileNameLengthInChars;
    TCHAR         ShortFileName[14];
    TCHAR         FileName[MAX_PATH];
    TCHAR *       Extension;
} SDIR_DIRENT, *PSDIR_DIRENT;

typedef BOOL (WINAPI * DISABLE_WOW_REDIRECT_FN)(PVOID*);
typedef HANDLE (WINAPI * FIND_FIRST_STREAM_FN)(LPCTSTR, DWORD, PSDIR_WIN32_FIND_STREAM_DATA, DWORD);
typedef BOOL (WINAPI * FIND_NEXT_STREAM_FN)(HANDLE, PSDIR_WIN32_FIND_STREAM_DATA);
typedef DWORD (WINAPI * GET_COMPRESSED_FILE_SIZE_FN)(LPCTSTR, LPDWORD);
typedef BOOL (WINAPI * GET_DISK_FREE_SPACE_EX_FN)(LPCTSTR, PLARGE_INTEGER, PLARGE_INTEGER, PLARGE_INTEGER);
typedef BOOL (WINAPI * GET_FILE_INFORMATION_BY_HANDLE_EX_FN)(HANDLE, ULONG, PVOID, DWORD);

typedef DWORD (* SDIR_METADATA_WIDTH_FN)(PSDIR_FMTCHAR, WORD, PSDIR_DIRENT);
typedef DWORD (* SDIR_COMPARE_FN)(PSDIR_DIRENT, PSDIR_DIRENT);
typedef DWORD (* SDIR_COLLECT_FN)(PSDIR_DIRENT, PWIN32_FIND_DATA, LPTSTR);
typedef DWORD (* SDIR_GENERATE_FROM_STRING_FN)(PSDIR_DIRENT, LPTSTR);

typedef struct _SDIR_COMPARE {
    SDIR_COMPARE_FN CompareFn;
    DWORD           CompareBreakCondition;
    DWORD           CompareInverseCondition;
} SDIR_COMPARE, *PSDIR_COMPARE;


#define SDIR_OPT_COLLECT          (0x01)
#define SDIR_OPT_DISPLAY          (0x02)
#define SDIR_OPT_ALLOW_DISPLAY    (0x04)
#define SDIR_OPT_ALLOW_SORT       (0x08)

typedef struct _SDIR_OPTS {
    TCHAR           SubDirWalk[MAX_PATH];

    TCHAR           ParentName[MAX_PATH];
    DWORD           ParentNameLength;

    DWORD           OsVersion;
    DWORD           ConsoleHeight;
    DWORD           ConsoleBufferWidth;
    DWORD           ConsoleWidth;
    DWORD           MetadataWidth;

    UCHAR           EnableAllocationSize;
    UCHAR           EnableCompressedFileSize;
    UCHAR           EnableCreateDate;
    UCHAR           EnableCreateTime;
    UCHAR           EnableFileAttributes;
    UCHAR           EnableFileExtension;
    UCHAR           EnableFileId;
    UCHAR           EnableFileName;
    UCHAR           EnableFileSize;
    UCHAR           EnableLinkCount;
    UCHAR           EnableNamedStreams;
    UCHAR           EnableOsVersion;
    UCHAR           EnableReparseTag;
    UCHAR           EnableShortName;
    UCHAR           EnableStreamCount;
    UCHAR           EnableSummary;
    UCHAR           EnableUsn;
    UCHAR           EnableVersion;
    UCHAR           EnableWriteDate;
    UCHAR           EnableWriteTime;

    BOOLEAN         EnablePause;
    BOOLEAN         EnableNameTruncation;
    BOOLEAN         OutputIsConsole;
    BOOLEAN         OutputIsUnicode;
    BOOLEAN         Cancelled;

    WORD            PreviousAttributes;

    LPTSTR          NewLine;

    DWORD           CurrentSort;
    SDIR_COMPARE    Sort[8];

    FIND_FIRST_STREAM_FN                 FindFirstStreamW;
    FIND_NEXT_STREAM_FN                  FindNextStreamW;
    GET_COMPRESSED_FILE_SIZE_FN          GetCompressedFileSize;
    GET_DISK_FREE_SPACE_EX_FN            GetDiskFreeSpaceEx;
    GET_FILE_INFORMATION_BY_HANDLE_EX_FN GetFileInformationByHandleEx;

} SDIR_OPTS, *PSDIR_OPTS;

typedef struct _SDIR_SUMMARY {
    ULONG           NumFiles;
    ULONG           NumDirs;
    SDIR_FILESIZE   TotalSize;
    SDIR_FILESIZE   CompressedSize;
    LARGE_INTEGER   FreeSize;
    LARGE_INTEGER   VolumeSize;
} SDIR_SUMMARY, *PSDIR_SUMMARY;

typedef struct _SDIR_OPT {
    DWORD                        EnableOffset;
    TCHAR                        Switch[3];
    UCHAR                        Default;
    SDIR_METADATA_WIDTH_FN       WidthFn;
    SDIR_COLLECT_FN              CollectFn;
    SDIR_COMPARE_FN              CompareFn;
    SDIR_COMPARE_FN              BitwiseCompareFn;
    SDIR_GENERATE_FROM_STRING_FN GenerateFromStringFn;
    CHAR                         Help[20];
} SDIR_OPT, *PSDIR_OPT;

#define SDIR_LESS_THAN    0
#define SDIR_EQUAL        1
#define SDIR_GREATER_THAN 2

#define SDIR_NOT_EQUAL    0

typedef struct _SDIR_EXEC {
    DWORD EnableOffset;
    SDIR_METADATA_WIDTH_FN Function;
} SDIR_EXEC, *PSDIR_EXEC;

extern DWORD SdirAllocatedDirents;
extern PSDIR_DIRENT SdirDirCollection;
extern PSDIR_DIRENT * SdirDirSorted;
extern ULONG SdirDirCollectionCurrent;
extern ULONG SdirDirCollectionLongest;
extern ULONG SdirDirCollectionTotalNameLength;

extern PSDIR_OPTS Opts;
extern PSDIR_SUMMARY Summary;
extern const SDIR_OPT SdirOptions[];
extern const SDIR_EXEC SdirExec[];
extern const CHAR SdirAttributeDefaultApplyString[];

//
//  Functions from display.c
//

BOOL
SdirWrite (
    __in_ecount(count) PSDIR_FMTCHAR str,
    __in DWORD count
    );

BOOL
SdirPasteStrAndPad (
    __out_ecount(padsize) PSDIR_FMTCHAR str,
    __in_ecount(count) LPTSTR src,
    __in WORD attr,
    __in DWORD count,
    __in DWORD padsize
    );

BOOL
SdirPressAnyKey ();

BOOL
SdirWriteStringWithAttribute (
    __in LPCTSTR str,
    __in WORD DefaultAttribute
    );

BOOL
SdirDisplayError (
    __in LONG ErrorCode,
    __in LPCTSTR Prefix
    );


#define SdirWriteString(str) SdirWriteStringWithAttribute(str, SDIR_DEFAULT_COLOR)

#define SdirPasteStr(str, src, attr, count) \
    SdirPasteStrAndPad(str, src, attr, count, count)

//
//  Functions from callbacks.c
//

DWORD
SdirGetNumSdirOptions();

DWORD
SdirGetNumSdirExec();

ULONG
SdirDisplaySummary(
    __in WORD DefaultAttributes
    );

BOOL
SdirCollectSummary(
    __in PSDIR_DIRENT Entry
    );

ULONG
SdirDisplayShortName (
    PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PSDIR_DIRENT Entry
    );

DWORD
SdirCompareFileName (
    __in PSDIR_DIRENT Left,
    __in PSDIR_DIRENT Right
    );

//
//  Functions from color.c
//

BOOL
SdirParseAttributeApplyString();

BOOL
SdirApplyAttribute(
    __in PSDIR_DIRENT DirEnt,
    __out WORD *Attribute
    );

//
//  Functions from init.c
//

BOOL
SdirInit(
    __in int argc,
    __in_ecount(argc) LPTSTR argv[]
    );

//
//  Functions from sdir.c
//

DWORD
SdirGetNumAttrPairs();

//
//  Functions from usage.c
//

BOOL
SdirUsage();

//
//  Functions from utils.c
//

DWORD
SdirCompareLargeInt (
    __in PULARGE_INTEGER Left,
    __in PULARGE_INTEGER Right
    );

DWORD
SdirCompareString (
    __in LPCTSTR Left,
    __in LPCTSTR Right
    );

DWORD
SdirCompareDate (
    __in LPSYSTEMTIME Left,
    __in LPSYSTEMTIME Right
    );

DWORD
SdirCompareTime (
    __in LPSYSTEMTIME Left,
    __in LPSYSTEMTIME Right
    );

DWORD
SdirStringToNum32(
    __in TCHAR * str,
    __out_opt TCHAR ** endptr
    );

LARGE_INTEGER
SdirStringToNum64(
    __in TCHAR * str,
    __out_opt TCHAR ** endptr
    );

LARGE_INTEGER
SdirStringToFileSize(
    __in TCHAR * str
    );

BOOL
SdirStringToDate(
    __in TCHAR * str,
    __out PSYSTEMTIME date
    );

BOOL
SdirStringToTime(
    __in TCHAR * str,
    __out PSYSTEMTIME date
    );

ULONG
SdirDisplayGenericSize(
    __out_ecount(6) PSDIR_FMTCHAR Buffer,
    __in WORD Attributes,
    __in PLARGE_INTEGER GenericSize
    );

ULONG
SdirDisplayHex64 (
    __out_ecount(18) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in PLARGE_INTEGER Hex
    );

ULONG
SdirDisplayHex32 (
    __out_ecount(9) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in DWORD Hex
    );

ULONG
SdirDisplayFileDate (
    __out_ecount(11) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in SYSTEMTIME * Time
    );

ULONG
SdirDisplayFileTime (
    __out_ecount(9) PSDIR_FMTCHAR Buffer,
    __in WORD DefaultAttributes,
    __in SYSTEMTIME * Time
    );

#define SdirAnsiToUnicode(u, l, a) _stprintf_s(u, l, _T("%hs"), a)

// vim:sw=4:ts=4:et:
